# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for OpenSpiel 'Repeated' PokerkitWrapper."""

import itertools
import random

from absl.testing import absltest
from absl.testing import parameterized
import pokerkit

from open_spiel.python.games import pokerkit_wrapper
from open_spiel.python.games import repeated_pokerkit
import pyspiel

ACTION_CHECK_OR_CALL = pokerkit_wrapper.ACTION_CHECK_OR_CALL
ACTION_FOLD = pokerkit_wrapper.ACTION_FOLD

Card = pokerkit.Card
ACE = pokerkit.Rank.ACE
DEUCE = pokerkit.Rank.DEUCE
KING = pokerkit.Rank.KING
SEVEN = pokerkit.Rank.SEVEN
SIX = pokerkit.Rank.SIX
SPADE = pokerkit.Suit.SPADE
CLUB = pokerkit.Suit.CLUB
HEART = pokerkit.Suit.HEART
DIAMOND = pokerkit.Suit.DIAMOND

# E.g.
# [
#   {variant: "NoLimitTexasHoldem"}
#   {variant: "FixedLimitTexasHoldem"}
#   ...
# ]
_ALL_VARIANTS = [
    {"variant": variant}
    for variant in pokerkit_wrapper.SUPPORTED_VARIANT_MAP.keys()
]
# E.g.
# [
#   {variant: "NoLimitTexasHoldem", reset_stacks: True},
#   {variant: "NoLimitTexasHoldem", reset_stacks: False},
#   {variant: "FixedLimitTexasHoldem", reset_stacks: True},
#   {variant: "FixedLimitTexasHoldem", reset_stacks: False},
#   ...
# ]
_ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS = [
    params | {"reset_stacks": reset_stacks}
    for params, reset_stacks in itertools.product(_ALL_VARIANTS, [True, False])
]


class RepeatedPokerkitTest(parameterized.TestCase):
  """Test the OpenSpiel game 'Repeated Pokerkit'."""

  def _load_repeated_pokerkit_game_with_variant(self, variant, params):
    """Loads a repeated_pokerkit game with variant-specific param filtering."""
    use_bring_in = variant in pokerkit_wrapper.VARIANT_PARAM_USAGE["bring_in"]
    use_blinds = (
        variant
        in pokerkit_wrapper.VARIANT_PARAM_USAGE["raw_blinds_or_straddles"]
    )
    use_fixed_bet_sizes = (
        variant in pokerkit_wrapper.VARIANT_PARAM_USAGE["small_bet"]
    )
    if use_fixed_bet_sizes:
      self.assertIn(variant, pokerkit_wrapper.VARIANT_PARAM_USAGE["big_bet"])

    if not use_bring_in:
      params["pokerkit_game_params"].pop("bring_in")
    if not use_blinds:
      self.assertTrue(use_bring_in and use_fixed_bet_sizes)
      params["pokerkit_game_params"].pop("blinds")
      params["rotate_dealer"] = False
    if not use_fixed_bet_sizes:
      self.assertTrue(use_blinds and not use_bring_in)
      params["pokerkit_game_params"].pop("small_bet")
      params["pokerkit_game_params"].pop("big_bet")

    return pyspiel.load_game("python_repeated_pokerkit", params)

  def test_parse_blind_schedule(self):
    blind_schedule_str = "1,1,2;2,2,3"
    blind_schedule = repeated_pokerkit.parse_blind_schedule(blind_schedule_str)
    self.assertLen(blind_schedule, 2)
    self.assertEqual(blind_schedule[0].num_hands, 1)
    self.assertEqual(blind_schedule[0].small_blind, 1)
    self.assertEqual(blind_schedule[0].big_blind, 2)
    self.assertEqual(blind_schedule[1].num_hands, 2)
    self.assertEqual(blind_schedule[1].small_blind, 2)
    self.assertEqual(blind_schedule[1].big_blind, 3)

  def test_random_sim_cash_game_headsup(self):
    # Note the reset_stacks=True
    game = pyspiel.load_game(
        "python_repeated_pokerkit(max_num_hands=20,reset_stacks=True,"
        "rotate_dealer=True,pokerkit_game_params=python_pokerkit_wrapper())"
    )
    pyspiel.random_sim_test(game, num_sims=5, serialize=False, verbose=True)

  def test_random_sim_tournament_headsup(self):
    # Note the reset_stacks=False
    game = pyspiel.load_game(
        "python_repeated_pokerkit(max_num_hands=20,reset_stacks=False,"
        "rotate_dealer=True,pokerkit_game_params=python_pokerkit_wrapper())"
    )
    pyspiel.random_sim_test(game, num_sims=5, serialize=False, verbose=True)

  @parameterized.parameters(*_ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS)
  def test_random_sim_all_variants_cash_and_tournament_headsup(
      self,
      variant,
      # True -> cash game, False -> tournament
      reset_stacks,
  ):
    params = {
        "max_num_hands": 10,
        "reset_stacks": reset_stacks,
        "rotate_dealer": True,
        "pokerkit_game_params": {
            "name": "python_pokerkit_wrapper",
            "variant": variant,
            "num_players": 2,
            "blinds": "2 4",
            "bring_in": 4,
            "small_bet": 8,
            "big_bet": 16,
            "stack_sizes": "64 64",
        },
    }
    game = self._load_repeated_pokerkit_game_with_variant(variant, params)
    # NOTE: Consider carefully before increasing num_sims any further. As of
    # 2025, num_sims=3 is already pushing it (and e.g. num_sims=5 takes ~50%
    # longer than num_sims=3 does).
    pyspiel.random_sim_test(game, num_sims=3, serialize=False, verbose=False)

  @parameterized.parameters(*_ALL_VARIANTS_CASH_AND_TOURNAMENT_PARAMS)
  def test_random_sim_all_variants_cash_and_tournament_6max(
      self,
      variant,
      # True -> cash game, False -> tournament
      reset_stacks,
  ):
    params = {
        "max_num_hands": 3,
        "reset_stacks": reset_stacks,
        "rotate_dealer": True,
        "pokerkit_game_params": {
            "name": "python_pokerkit_wrapper",
            "variant": variant,
            "num_players": 6,
            "blinds": "2 4",
            "bring_in": 4,
            "small_bet": 8,
            "big_bet": 16,
            "stack_sizes": "30 30 30 30 30 30",
        },
    }
    game = self._load_repeated_pokerkit_game_with_variant(variant, params)
    # NOTE: Consider carefully before increasing num_sims any further. As of
    # 2025, num_sims=2 is already pushing it (and e.g. num_sims=3 takes ~50%
    # longer than num_sims=2 does).
    pyspiel.random_sim_test(game, num_sims=2, serialize=False, verbose=False)

  def test_blind_schedule_progression(self):
    blind_schedule_str = "1,10,20;2,20,40;1,50,100;1,1000,1000"
    params = {
        "max_num_hands": 100,
        "reset_stacks": False,
        "rotate_dealer": True,
        "blind_schedule": blind_schedule_str,
        "pokerkit_game_params": {
            "name": "python_pokerkit_wrapper",
            "num_players": 2,
            "blinds": "1 2",  # Initial blinds should be overridden
            "stack_sizes": "1000 1000",
        },
    }
    game = pyspiel.load_game("python_repeated_pokerkit", params)
    state = game.new_initial_state()

    # TODO: b/444333187 - consider refactoring to be "DAMP" instead of "DRY".
    expected_blinds = [
        (10, 20),  # Hand 0
        (20, 40),  # Hand 1
        (20, 40),  # Hand 2
        (50, 100),  # Hand 3
        (1000, 1000),  # Hands 4+
    ]
    seen_hands = set()
    current_hand = state._hand_number
    self.assertEqual(state._hand_number, 0)
    while not state.is_terminal():
      self.assertEqual(state._hand_number, current_hand)
      if current_hand not in seen_hands:
        seen_hands.add(current_hand)
        small_blind, big_blind = (
            expected_blinds[-1]
            if current_hand >= len(expected_blinds)
            else expected_blinds[current_hand]
        )
        self.assertEqual(
            state._small_blind,
            small_blind,
        )
        self.assertEqual(
            state._big_blind,
            big_blind,
        )

      if state.is_chance_node():
        action = random.choice([o for o, p in state.chance_outcomes()])
      else:
        legal_actions = state.legal_actions()
        # Choose to check/call in all situations to ensure we reach the final
        # blind schedule level (at which point everyone will be forced all-in).
        if ACTION_CHECK_OR_CALL in legal_actions:
          action = ACTION_CHECK_OR_CALL
        else:
          action = random.choice(state.legal_actions())

      state.apply_action(action)
      # apply_action handles the transition to the next hand when the
      # pokerkit_wrapper state is terminal.
      self.assertBetween(state._hand_number, current_hand, current_hand + 1)
      current_hand = state._hand_number

  def test_stacks_carried_over(self):
    params = {
        "max_num_hands": 3,
        "reset_stacks": False,
        "rotate_dealer": True,
        "pokerkit_game_params": {
            "name": "python_pokerkit_wrapper",
            "num_players": 3,
            "blinds": "10 20",
            "stack_sizes": "100 100 100",  # Players 0 1 2: SB BB BTN
        },
    }
    game = pyspiel.load_game("python_repeated_pokerkit", params)
    state = game.new_initial_state()

    # Hand 0: P0 will win, P1 will bust, P2 will not be affected:

    self.assertEqual(state._hand_number, 0)
    self.assertEqual(state._stacks, [100, 100, 100])
    # Play a hand where P0 wins all of P1's chips and none of P2's chips.
    # P0 (SB) posts 10, P1 (BB) posts 20

    # First hole card (P0, then P1, then P2)
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(ACE, SPADE)]
    )
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(DEUCE, SPADE)]
    )
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(KING, HEART)]
    )

    # Second hole card (P0, then P1, then P2)
    # P0 has pocket Aces, P1 pockets Twos, P2 pocket Kings.
    state.apply_action(state.pokerkit_wrapper_game.card_to_int[Card(ACE, CLUB)])
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(DEUCE, CLUB)]
    )
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(KING, SPADE)]
    )

    # P2 folds preflop
    state.apply_action(ACTION_FOLD)
    # P0 calls to 20 (10 additional), P1 checks
    state.apply_action(ACTION_CHECK_OR_CALL)
    state.apply_action(ACTION_CHECK_OR_CALL)
    # Flop
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, SPADE)]
    )
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, HEART)]
    )
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(SEVEN, DIAMOND)]
    )
    # P0 bets 80 (all-in), P1 calls
    state.apply_action(80)
    state.apply_action(ACTION_CHECK_OR_CALL)
    # Turn/River
    state.apply_action(state.pokerkit_wrapper_game.card_to_int[Card(SIX, CLUB)])
    state.apply_action(
        state.pokerkit_wrapper_game.card_to_int[Card(SIX, DIAMOND)]
    )
    assert not state.is_terminal()

    # P0 wins the pot. P0 stack: 100 + 100 = 200. P1 stack: 100 - 100 = 0.
    # P2 stack is unchanged. It's now the next hand and only P0 and P2 are still
    # playing (ie in the underlying pokerkit game).
    self.assertEqual(state._hand_number, 1)
    self.assertEqual(state._stacks, [200, 0, 100])
    # Button rotates from P2 to P0 since P0 is still active. But, this is heads
    # up now, meaning:
    # - P0 (former SB) now SB/BTN: 200 - 10 => 190
    # - P2 (former BTN) now BB: 100 - 20 => 80
    self.assertEqual(
        state.pokerkit_wrapper_state._wrapped_state.stacks,
        # NOTE: This is the pokerkit's stacks, meaning that the
        # last element is the BTN (or in this case, SB/BTN). So
        # this makes sense given the rotation from the prior hand.
        [80, 190],
    )

  def test_eliminates_players_at_exactly_zero_chips(self):
    params = {
        "max_num_hands": 3,
        "reset_stacks": False,
        "rotate_dealer": True,
        "pokerkit_game_params": {
            "name": "python_pokerkit_wrapper",
            "num_players": 3,
            "blinds": "10 20",
            "stack_sizes": "100 100 100",  # Players 0 1 2: SB BB BTN
        },
    }
    game = pyspiel.load_game("python_repeated_pokerkit", params)
    state = game.new_initial_state()

    # Deal each player two random cards (choice doesn't matter)
    # 3 players * 2 hole cards => 6 chance nodes
    for _ in range(6):
      action = random.choice([o for o, _ in state.chance_outcomes()])
      state.apply_action(action)

    # P3 BTN raises to *almost* all-in (1 chip behind)
    state.apply_action(99)
    # P1 SB Shoves all-in
    state.apply_action(100)
    # P2 BB folds => 80 chips left
    state.apply_action(ACTION_FOLD)
    # BTN folds => 1 chip left (should not bust). BB wins the pot and the next
    # hand starts.
    self.assertEqual(state._hand_number, 0)
    state.apply_action(ACTION_FOLD)
    self.assertEqual(state._hand_number, 1)
    self.assertEqual(state._stacks, [219, 80, 1])

    # Ensure *three* players here are still active, not just two.
    self.assertEqual(state.num_players(), 3)
    self.assertEqual(state._num_active_players, 3)
    # Similarly: 3 players * 2 hole cards => 6 chance nodes
    for _ in range(6):
      self.assertTrue(state.is_chance_node())
      self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)
      action = random.choice([o for o, _ in state.chance_outcomes()])
      state.apply_action(action)

      # Double check again that nothing unexpectedly changes out from underneath
      # us here as we apply actions.
      self.assertEqual(state.num_players(), 3)
      self.assertEqual(state._num_active_players, 3)

    # P1 BTN bets 40
    state.apply_action(40)
    # P2 SB calls
    state.apply_action(ACTION_CHECK_OR_CALL)
    # P3 BB was already all-in due to having less than one Big Blind, so
    # gameplay should immediately procede to dealing the flop despite there
    # being 3 players.
    self.assertTrue(state.is_chance_node())

    # Definitely excessive, but triple checking this again just to be certain.
    self.assertEqual(state.num_players(), 3)
    self.assertEqual(state._num_active_players, 3)


# TODO: b/444333187 - Add tests for the following behaviors:
# - blind schedule becoming higher than all players' remaining stacks. (This was
#   a known issue previously, though may be fixed now that we no longer are
#   eliminating players with less than one Big Blind.)
# - reaching a low max_num_hands limit
# - multiple players busting out (especially edge cases)
# - returns()'s values being correct when reaching terminal state


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3

"""A Python wrapper for playing repeated games of poker using Pokerkit.

This module defines the `RepeatedPokerkit` game, which allows for playing
multiple hands of poker within a single game episode. It supports features like
blind schedules for tournament play and stack management across hands.
"""

import copy
import dataclasses

from absl import logging
import numpy as np
import pokerkit

from open_spiel.python.games import pokerkit_wrapper
import pyspiel


dataclass = dataclasses.dataclass
DEFAULT_NUM_HANDS = 1
INACTIVE_PLAYER_SEAT = -1
INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE = -1


@dataclass
class BlindLevel:
  num_hands: int
  small_blind: int
  big_blind: int


@dataclass
class BetSizeLevel:
  num_hands: int
  small_bet_size: int
  big_bet_size: int


@dataclass
class BringInLevel:
  num_hands: int
  bring_in: int


# WARNING: pickle serialization/deserialization of this does not work properly
# (specifically, when handling the nested kGame for the python_pokerkit_wrapper
# inputs). In meantime we have disabled testing the GameType for this game.
# TODO: b/445192003 - Investigate + resolve the nested kGame issue.
_DEFAULT_PARAMS = {
    "pokerkit_game_params": {
        "name": "python_pokerkit_wrapper",
        # See pokerkit_wrapper_test.py for other options, e.g.
        # "variant": "NoLimitTexasHoldem",
        # "num_players": 2,
        # "blinds": "50 100",
        # "stack_sizes": "20000 20000",
        # ...
    },
    "max_num_hands": DEFAULT_NUM_HANDS,
    "reset_stacks": False,
    "rotate_dealer": True,
    "blind_schedule": "",
    "bet_size_schedule": "",
    "bring_in_schedule": "",
    "first_button_player": (
        # Will be changed to the last player if not overridden.
        -1
    ),
}

_GAME_TYPE = pyspiel.GameType(
    short_name="python_repeated_pokerkit",
    long_name="Python Repeated Pokerkit",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=10,  # Arbitrarily chosen to match universal_poker
    min_num_players=2,  # Arbitrarily chosen to match universal_poker
    # Not yet supported. Instead use the observation string.
    provides_information_state_string=False,
    provides_observation_string=True,
    # Tensors are not yet supported. Instead use the corresponding strings.
    provides_information_state_tensor=False,
    provides_observation_tensor=False,
    # WARNING: pickle serialization/deserialization of our _DEFAULT_PARAMS does
    # not work properly. In meantime we have disabled testing the GameType for
    # this game.
    # TODO: b/445192003 - Investigate + resolve the serialization issue.
    parameter_specification=_DEFAULT_PARAMS,
)


def _split_level_schedule_string(
    schedule_str: str, expected_level_parts: int
) -> list[list[str]]:
  """Splits a level schedule string into a 2-D list of strings.

  Parsed schedule strings are of the form
    <level_1>;<level_2>;...;<level_n>
  where each level is of the form
    <substring_1>,<substring_2>,<substring_3>

  Args:
    schedule_str: The schedule string to split. The format is a
      semicolon-separated list of inner levels, where each level is a
      comma-separated tuple of `expected_level_parts` substrings.
    expected_level_parts: The expected number of comma-separated substrings
      within each level.

  Returns:
    A 2-D list of strings, where the first dimension corresponds to the levels
    and the second dimension corresponds to the substrings within each level.
  """
  if not schedule_str:
    return []
  levels_str = schedule_str.removesuffix(";").split(";")
  levels = [level.split(",") for level in levels_str]

  for level in levels:
    if len(level) != expected_level_parts:
      raise ValueError(
          f"Invalid schedule string: {schedule_str}. Expected format is"
          " <level_1>;<level_2>;...;<level_n> where each level is a"
          f" comma-separated tuple of {expected_level_parts} substrings."
      )
  return levels


def parse_blind_schedule(blind_schedule_str: str) -> list[BlindLevel]:
  """Parses a blind schedule string into a list of BlindLevel objects.

  Port of ParseBlindSchedule from repeated_poker.cc.

  Parses blind schedule string of the form
    <blind_level_1>;...;<blind_level_n>
  where each blind level is of the form
    <num_hands>,<small_blind>,<big_blind>

  Args:
    blind_schedule_str: A string specifying the blind schedule. The format is a
      semicolon-separated list of blind levels, where each level is a
      comma-separated tuple of `num_hands`, `small_blind`, and `big_blind`.

  Returns:
    A list of BlindLevel objects parsed from the input string.
  """
  if not blind_schedule_str:
    return []

  blind_levels = []
  for level_parts in _split_level_schedule_string(blind_schedule_str, 3):
    num_hands, small_blind, big_blind = [int(p) for p in level_parts]
    blind_levels.append(
        BlindLevel(
            num_hands,
            small_blind,
            big_blind,
        )
    )
  return blind_levels


def parse_bet_size_schedule(bet_size_schedule_str: str) -> list[BetSizeLevel]:
  """Parses a bet-size schedule string into a list of BetSizeLevel objects.

  Parsed bet-size schedule strings are of the form
    <bet_size_level_1>;...;<bet_size_level_n>
  where each bet-size level is of the form
    <num_hands>,<small_bet_size>,<big_bet_size>

  Args:
    bet_size_schedule_str: A string specifying the bet-size schedule. The format
      is a semicolon-separated list of bet-size levels, where each level is a
      comma-separated tuple of `num_hands`, `small_bet_size`, and
      `big_bet_size`.

  Returns:
    A list of BetSizeLevel objects parsed from the input string.
  """
  if not bet_size_schedule_str:
    return []

  bet_sizes = []
  for level_parts in _split_level_schedule_string(bet_size_schedule_str, 3):
    num_hands, small_bet_size, big_bet_size = [int(p) for p in level_parts]
    bet_sizes.append(BetSizeLevel(num_hands, small_bet_size, big_bet_size))
  return bet_sizes


def parse_bring_in_schedule(bring_in_schedule_str: str) -> list[BringInLevel]:
  """Parses a bring-in schedule string into a list of BringInLevel objects.

  Parsed bring-in schedule strings are of the form
    <bring_in_level_1>;...;<bring_in_level_n>
  where each bring-in level is of the form
    <num_hands>,<bring_in>

  Args:
    bring_in_schedule_str: A string specifying the bring-in schedule. The format
      is a semicolon-separated list of bring-in levels, where each level is a
      comma-separated tuple of `num_hands` and `bring_in`.

  Returns:
    A list of BringInLevel objects parsed from the input string.
  """
  if not bring_in_schedule_str:
    return []
  bring_ins = []
  for level_parts in _split_level_schedule_string(bring_in_schedule_str, 2):
    num_hands, bring_in = [int(p) for p in level_parts]
    bring_ins.append(BringInLevel(num_hands, bring_in))
  return bring_ins


# TODO: b/437724266 - Refactor to remove direct usage of protected fields and
# then remove the above disable. (For now we're intentionally using them to
# ensure that this code mimics the RepeatedPoker game's behavior as closely as
# possible during the initial commit.)
#
# pylint: disable=protected-access


class RepeatedPokerkit(pyspiel.Game):
  """Wrapper around PokerkitWrapper for playing multiple hands in one episode.

  Wrapper around PokerkitWrapper (i.e.  'double-wrapped' Pokerkit) for playing
  multiple hands within the same game episode. This enables simulating both cash
  games and tournaments.

  TODO: b/444333187 - does not fully support games besides NoLimitTexasHoldem
  yet. See the TODOs below for more details.

  Parameters:
    "pokerkit_game_params":  (required)
        Specifies the underlying pokerkit game to begin play. Note that the
        params will be updated to reflect the current state of the repeated
        game (e.g., number of players, blinds, etc.).
        The start of this string must be "python_pokerkit_wrapper(" or
        "python_pokerkit_wrapper_acpc_style(" and then end must be a closing
        paren ")".
    "max_num_hands": int (required)
        The maximum number of hands to play in the episode. This is a required
        parameter because it should be set deliberately in relation to the
        blind schedule when playing a tournament.
    "reset_stacks": bool (required)
        Whether to reset the stacks at the start of each hand. Required.
    "rotate_dealer": bool (optional, default=True)
        Whether to rotate the dealer at the start of each hand (which for
        pokerkit actually means rotating *the seats* so that the Button is the
        last stack provided to pokerkit). This defaults to true as it is always
        done in practice. NOTE: We will ignore this parameter being False if the
        resulting dealer player busts out when reset_stacks is False and there
        are still additional hands to play. I.e. in that situation the next
        not-yet-busted player will become the dealer going forwards.
    "blind_schedule": string (optional)
        Specifies the blind schedule for playing a tournament. The format is:
        <blind_level_1>;<blind_level_2>;...<blind_level_n> where each blind
        level is of the form <num_hands>,<small_blind>,<big_blind>. If play
        continues beyond the number of hands specified in the last blind level,
        the last blind level will continue to be used.
    "bet_size_schedule": string (optional)
        Specifies the bet-size schedule for playing a tournament. The format is:
        <bet_size_level_1>;<bet_size_level_2>;...<bet_size_level_n> where each
        bet-size level is of the form
        <num_hands>,<small_bet_size>,<big_bet_size>.  If play continues beyond
        the number of hands specified in the last bet-size level, the last
        bet-size level will continue to be used.
    "bring_in_schedule": string (optional)
        Specifies the bring-in schedule for playing a tournament. The format is:
        <bring_in_level_1>;<bring_in_level_2>;...<bring_in_level_n> where each
        bring-in level is of the form <num_hands>,<bring_in>. If play continues
        beyond the number of hands specified in the last bring-in level, the
        last bring-in level will continue to be used.
    "first_button_player": int (optional, default=-1)
        The player ID of the first player that will be the last seat, and
        therefore the Button, when performing seat assignments for the very
        first hand. (Pokerkit always assigns positions based on player order,
        with the last player corresponding to the button.)
        If left as the default value -1, this value will be overridden with the
        last player ID, effectively applying pokerkit's default interpretation
        onto the player IDs' order for the first hand. For example:
        - 1 in heads-up games (effectively making the BB the player with ID 0
          and the SB/BTN the player with ID 1)
        - 2 in 3-player games (effectively making the SB the player ID with ID
          0, the BB the player with ID 1, and the BTN the player with ID 2)
        - 5 in 6-player games (effectively making the SB the player ID with ID
          0, the BB the player with ID 1, the UTG player with ID 2 ... the BTN
          the player with ID 2)
        and so on.

  Returns are calculated by summing the returns for each hand.

  pylint: disable=g-bad-todo (copied from RepeatedPoker)
  TODO: jhtschultz - Support payout structures for tournaments.
  pylint: enable=g-bad-todo

  Note that this implementation imposes some slightly stricter assumptions on
  the game definition:
   1. Exactly two blinds.
   2. At least 2 rounds.
  Both of these are very standard assumptions for poker as played in practice.

  NOTE: we use simplified moving button rules. See
  https://en.wikipedia.org/wiki/Betting_in_poker#When_a_player_in_the_blinds_leaves_the_game
  This is common in online poker games generally speaking, and it is used here
  because the logic for remapping each hand to a gamedef becomes quite
  complex and highly error-prone when using dead button rules.

  NOTE: This implementation differs from RepeatedPoker in that it *does* support
  normal tournament rules for players with stack sizes less than one Big Blind.
  (Specifically: we allow players to stay in and play as per normal, rather than
  automatically eliminating them in such cases. This means that the total
  rewards should always add up to the number of total chips when the tournament
  started, as would be assumed for typical poker tournaments.)
  """

  def __init__(self, params=None):
    self.params = params
    self._pokerkit_game_params = {}
    self._max_num_hands = params.get("max_num_hands")
    if not self._max_num_hands or not isinstance(self._max_num_hands, int):
      raise ValueError(
          "max_num_hands must be an int >= 1. Got: %d" % self._max_num_hands
      )
    self._reset_stacks = params.get("reset_stacks")
    self._rotate_dealer = params.get("rotate_dealer")
    self._blind_schedule = params.get("blind_schedule")
    self._bet_size_schedule = params.get("bet_size_schedule")
    self._bring_in_schedule = params.get("bring_in_schedule")

    self._base_game_params = params.get("pokerkit_game_params")
    if not self._base_game_params:
      raise ValueError("pokerkit_game_params must be provided.")
    if self._base_game_params.get("name") not in [
        "python_pokerkit_wrapper",
        "python_pokerkit_wrapper_acpc_style",
    ]:
      raise ValueError(
          "pokerkit_game_params must parse to python_pokerkit_wrapper or"
          " python_pokerkit_wrapper_acpc_style game params. Instead got game"
          f" name {self._base_game_params.get('name')} inside "
          f" parsed params of: {self._base_game_params}"
      )
    # NOTE: Not used anywhere in the actual 'State' below! Just used to figure
    # out the basic game properties at the very very start of the repeated game
    # (e.g. num players, deck size, utility calculations, etc)
    #
    # Additionally, this may not have the seat rotation specified by the
    # first_button_player param (since we use properties of this game in order
    # to actually set that value in certain cases).
    self._base_game = pyspiel.load_game(
        pyspiel.game_parameters_to_string(self._base_game_params)
    )
    self._game_info = pyspiel.GameInfo(
        num_distinct_actions=self._calculate_num_distinct_actions(),
        max_chance_outcomes=self._base_game.deck_size,
        num_players=self._base_game.num_players(),
        min_utility=self._calculate_min_utility(),
        max_utility=self._calculate_max_utility(),
        utility_sum=0.0,
        max_game_length=self._calculate_max_game_length(),
    )

    provided_first_button_player = params.get("first_button_player")
    if (
        provided_first_button_player is None
        or provided_first_button_player < -1
        or provided_first_button_player >= self._base_game.num_players()
    ):
      raise ValueError(
          "first_button_player must be a valid player ID or the value -1. Got:"
          f" {provided_first_button_player}"
      )
    self._first_button_player = (
        provided_first_button_player
        if provided_first_button_player >= 0
        else self._base_game.num_players() - 1
    )

    super().__init__(_GAME_TYPE, self._game_info, self.params)

  def _calculate_num_distinct_actions(self):
    """Returns an upper bound on the number of distinct player actions.

    Port of RepeatedPokerGame::NumDistinctActions.

    The action space must be able to support all legal bet sizes. This remains
    constant in the case of resetting stacks, but otherwise will increase as
    stack sizes increases. Players can bet more than their opponent's stack
    size, so the upper bound is set to the total number of chips. This isn't a
    tight upper bound, but it's very close and should not matter for practical
    purposes.
    """
    if self._reset_stacks:
      return self._base_game.num_distinct_actions()
    else:
      return sum(self._base_game.stack_sizes)

  def _calculate_min_utility(self):
    """Returns an upper-bound (magnitude) on minimum utility for the game.

    Port of RepeatedPokerGame::MinUtility.

    Note: Unlike universal_poker, pokerkit_wrapper doesn't have fine-grained
    utility calculations based on e.g. fixed limit bet sizes, so this may be
    significantly less tight in some cases!
    """
    if self._reset_stacks:
      return self._base_game.min_utility() * self._max_num_hands
    else:
      return max(self._base_game.stack_sizes) * -1

  def _calculate_max_utility(self):
    """Returns an upper-bound on maximum utility for the game.

    Port of RepeatedPokerGame::MaxUtility.

    Note: Unlike universal_poker, pokerkit_wrapper doesn't have fine-grained
    utility calculations based on e.g. fixed limit bet sizes, so this will be
    significantly less tight!
    """
    if self._reset_stacks:
      return self._base_game.max_utility() * self._max_num_hands
    else:
      return sum(self._base_game.stack_sizes)

  def game_info(self):
    return self._game_info

  def _calculate_max_game_length(self):
    """Provides a (very rough) upper bound on the maximum game length."""
    # Warning: this may be too small if the effective stack sizes are being used
    # to determine the max game length in the future. Since later hands may be
    # longer than the initial hand's length, e.g. 1 chip vs 1000 chips heads up.
    return self._base_game.max_game_length() * self._max_num_hands

  def new_initial_state(self):
    return RepeatedPokerkitState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    return RepeatedPokerkitObserver(self, iig_obs_type, params)

  # Not yet actually supported. Providing this 'return hardcoded value'
  # placeholder implementation instead of throwing an error to avoid crashing
  # anything in OpenSpiel that doesn't properly respect the
  # provides_information_state_tensor=False setting above.
  def information_state_tensor_size(self) -> int:
    logging.warning(
        "information_state_tensor_size() is not yet supported for"
        " RepeatedPokerkit. Returning a default value of 1 for now."
    )
    return 1


class RepeatedPokerkitState(pyspiel.State):
  """Represents an _OpenSpiel_ 'state' for the RepeatedPokerkit game.

  Port of repeated_poker::RepeatedPokerState.

  As described by the name, this class indeed wraps a `pokerkit.State` object
  and provides the necessary interface for OpenSpiel's `pyspiel.State`.
  """

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    self._hand_number = 0
    self._is_terminal = False
    # Represents the stack sizes at the start of the current hand. To access
    # the stacks for the hand in progress, use the underlying
    # PokerkitWrapperState.
    self._stacks: list[int] = None
    self._dealer = pyspiel.PlayerId.INVALID
    self._seat_to_player: dict[int, int] = None
    self._player_to_seat: dict[int, int] = None
    self._small_blind = INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
    self._big_blind = INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
    self._small_bet_size = INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
    self._big_bet_size = INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
    self._bring_in = INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE

    self._wrapped_state_hand_histories: list[str] = []
    # 2-D list where each inner list has length _num_players
    self._hand_returns: list[list[float]] = [[0.0] * game.num_players()]
    self._player_to_seat = {}
    self._seat_to_player = {}

    super().__init__(game)

    # Load the underlying pokerkit wrapper game and state that we're going to
    # actually send actions / update whenever starting a new hand. Also, save
    # its parameters back out so that we can mutate it within the state without
    # modifying the original game params that were passed in to the
    # RepeatedPokerkit constructor.
    #
    # NOTE: Do not confuse with the ._base_game on the RepeatedPokerGame above!!
    # (They are the same type, but only this one is actually 'used' for real
    # gameplay purposes.)
    self.pokerkit_wrapper_game: pyspiel.Game = pyspiel.load_game(
        # NOTE: This is the one time we load the original params in
        # ._base_game_params within the RepeatedPokerkitState. Everywhere else
        # we're going to use the ._pokerkit_game_params that are set below.
        pyspiel.game_parameters_to_string(game._base_game_params)
    )
    self._pokerkit_game_params = self.pokerkit_wrapper_game.get_parameters() | {
        "name": game._base_game_params.get("name")
    }
    self.pokerkit_wrapper_state: pyspiel.State = (
        self.pokerkit_wrapper_game.new_initial_state()
    )

    # Initial setup logic
    self._num_active_players = self.pokerkit_wrapper_game.num_players()
    self._hand_number = 0
    assert game._max_num_hands >= 1

    players = range(game.num_players())
    self._stacks = [
        self.pokerkit_wrapper_state._wrapped_state.starting_stacks[p]
        for p in players
    ]
    self._seat_to_player = {i: i for i in players}
    self._player_to_seat = {i: i for i in players}

    # Pokerkit has no equivalent of firstPlayer, so no need to port most of the
    # remaining RepeatedPokerState constructor logic here. BUT, we do at least
    # need to initialize dealer position to allow for users to specify an
    # arbitrary starting seat rotation that doesn't depend on which player ID
    # happens to be first or last.
    #
    # NOTE: This is a player ID, not a seat number!
    self._dealer = game._first_button_player
    if self._dealer < 0:
      raise ValueError(
          "first_button_player must be a valid player ID. Got:"
          f" {game._first_button_player}"
      )
    if self._dealer >= self.num_players():
      raise ValueError(
          "first_button_player must be a valid player ID. Got:"
          f" {game._first_button_player}"
      )
    if self._dealer >= self._num_active_players:
      raise ValueError(
          "first_button_player was unexepctedly high; ID. Got:"
          f" {game._first_button_player} but there are only"
          f" {self._num_active_players} active players upon initialization."
      )

    self._blind_schedule_levels: list[BlindLevel] = parse_blind_schedule(
        game._blind_schedule
    )
    self._bet_size_schedule_levels: list[BetSizeLevel] = (
        parse_bet_size_schedule(game._bet_size_schedule)
    )
    self._bring_in_schedule_levels: list[BringInLevel] = (
        parse_bring_in_schedule(game._bring_in_schedule)
    )
    if game._blind_schedule and not self._blind_schedule_levels:
      raise ValueError("Failed to parse blind schedule.")
    if game._bet_size_schedule and not self._bet_size_schedule_levels:
      raise ValueError("Failed to parse bet-size schedule.")
    if game._bring_in_schedule and not self._bring_in_schedule_levels:
      raise ValueError("Failed to parse bring-in schedule.")

    variant = game._base_game_params.get("variant")
    uses_blinds = (
        # NOTE: Assumes that if not specified, the default pokerkit_wrapper game
        # variant is going to be a Texas Holdem style game with blinds.
        variant is None
        or variant
        in pokerkit_wrapper.VARIANT_PARAM_USAGE["raw_blinds_or_straddles"]
    )
    if uses_blinds:
      if self._blind_schedule_levels:
        self._small_blind = self._blind_schedule_levels[0].small_blind
        self._big_blind = self._blind_schedule_levels[0].big_blind
      elif self.pokerkit_wrapper_state._wrapped_state.blinds_or_straddles:
        # Identify the small and big blind from the underlying pokerkit game.
        # (When set, it's always a list of integers with length equal to the
        # number of players in the pokerkit game)
        blinds = self.pokerkit_wrapper_state._wrapped_state.blinds_or_straddles
        num_blinds = 0
        for blind in blinds:
          if blind > 0:
            num_blinds += 1
            if self._small_blind == INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE:
              self._small_blind = blind
            else:
              self._big_blind = max(self._small_blind, blind)
              self._small_blind = min(self._small_blind, blind)
        if (
            any([blind != 0 for blind in blinds])
            or len(blinds) != self.num_players()
        ):
          assert num_blinds == 2
          assert self._small_blind > 0
          assert self._big_blind > 0
      else:
        # Many game variants (correctly) have no blinds.
        pass

    if not uses_blinds and game._rotate_dealer:
      raise ValueError(
          "Attempting to rotate dealers in a poker variant that does not have "
          " a rotating Button position. This is not supported. Poker variant "
          f" was: {variant}"
      )

    betting_structure = (
        self.pokerkit_wrapper_state._wrapped_state.betting_structure
    )
    if self._bet_size_schedule_levels:
      if betting_structure != pokerkit.state.BettingStructure.FIXED_LIMIT:
        raise ValueError(
            "Attempting to use a bet-size schedule, but the underlying pokerkit"
            " game does not have a fixed-limit betting structure. Aborting to"
            " avoid setting bet sizes to inaccurate values."
        )
      self._small_bet_size = self._bet_size_schedule_levels[0].small_bet_size
      self._big_bet_size = self._bet_size_schedule_levels[0].big_bet_size
    elif betting_structure == pokerkit.state.BettingStructure.FIXED_LIMIT:
      self._small_bet_size = self.pokerkit_wrapper_state.get_game().params.get(
          "small_bet"
      )
      self._big_bet_size = self.pokerkit_wrapper_state.get_game().params.get(
          "big_bet"
      )
    else:
      # Other game variants (correctly) have no bet sizes.
      logging.info("Found no small_bet or big_bet.")
      pass
    if self._bring_in_schedule_levels:
      if betting_structure != pokerkit.state.BettingStructure.FIXED_LIMIT:
        raise ValueError(
            "Attempting to use a bring-in schedule, but the underlying pokerkit"
            " game does not have a fixed-limit betting structure. Aborting to"
            " avoid setting bring-in to inaccurate values."
        )
      self._bring_in = self._bring_in_schedule_levels[0].bring_in
    elif betting_structure == pokerkit.state.BettingStructure.FIXED_LIMIT:
      self._bring_in = self.pokerkit_wrapper_state._wrapped_state.bring_in
    else:
      # Other game variants (correctly) have no bring-in.
      logging.info("Found no bring_in.")
      pass

    self.update_seat_rotation()
    self.update_blinds_bet_sizes_and_or_bring_in()
    self.update_pokerkit_wrapper()
    return

  def update_stacks(self):
    """Updates player stacks based on the outcome of the last hand.

    Port of RepeatedPokerState::UpdateStacks.

    Raises:
      RuntimeError: If `update_stacks` is called when the last pokerkit
        operation was not a `ChipsPulling` operation, which indicates the hand
        is not in a state where stacks can be safely updated.
    """
    if self.get_game()._reset_stacks:
      return
    # Unlike ACPC, pokerkit doesn't include chips in the payoff if they've
    # actively been bet / are in the pot. So, we need to be careful not to start
    # 'leaking' out chips from players' stacks if this is called at points where
    # the hand isn't actually over yet (e.g. start of the hand / any time the
    # hand is in progress)
    if all([
        isinstance(op, pokerkit.BlindOrStraddlePosting)
        for op in self.pokerkit_wrapper_state._wrapped_state.operations
    ]):
      return

    if not isinstance(
        self.pokerkit_wrapper_state._wrapped_state.operations[-1],
        pokerkit.ChipsPulling,
    ):
      raise RuntimeError(
          "Attempting to update stacks, but last pokerkit operation was not a"
          " ChipsPulling operation. Aborting to avoid setting stacks to"
          " inaccurate values (as there could be chips in the pot that we would"
          " effectively 'leak'). Operations was:",
          self.pokerkit_wrapper_state._wrapped_state.operations,
      )

    for player in range(self.num_players()):
      seat = self._player_to_seat[player]
      if seat != INACTIVE_PLAYER_SEAT:
        self._stacks[
            player
        ] += self.pokerkit_wrapper_state._wrapped_state.payoffs[seat]

  def update_seat_assignments_unrotated(self):
    """Updates our seat assignments for the next hand, without rotating seats.

    Port of RepeatedPokerState::UpdateSeatAssignments.

    As per the name, this doesn't handle the seat rotations necessary for
    moving the Button in pokerkit games. See instead
    `update_dealer_and_rotate_seats` below for more details.
    """
    if self.get_game()._reset_stacks:
      return

    self._player_to_seat = {}
    self._seat_to_player = {}
    next_open_seat = 0
    for player in range(self.num_players()):
      # Worst case the player should have 0 chips left; if they have a negative
      # number then there was likely a major bug in the game logic elsewhere.
      assert(self._stacks[player] >= 0)
      if self._stacks[player] == 0:
        self._player_to_seat[player] = INACTIVE_PLAYER_SEAT
      else:
        self._player_to_seat[player] = next_open_seat
        self._seat_to_player[next_open_seat] = player
        next_open_seat += 1
    self._num_active_players = next_open_seat

  def update_dealer(self):
    """Updates the dealer value. NOTE: does NOT rotate seats based on it!

    This is partially a port of RepeatedPokerState::UpdateDealer, with the
    exception of seat rotation logic (which we handle separately due to
    pokerkit differing from ACPC in how it works).

    Raises:
      RuntimeError: If an infinite loop is detected when rotating the dealer.
    """
    if (
        not self.get_game()._rotate_dealer
        # We MUST rotate the dealer position if the current dealer has busted
        # out (even if rotate_dealer is False). Otherwise we will be unable to
        # compute a proper seat rotation later on.
        and self._player_to_seat[self._dealer] != INACTIVE_PLAYER_SEAT
    ):
      return

    # Make sure our seat <-> player mappings are consistent prior to rotating.
    for player, seat in self._player_to_seat.items():
      if seat != INACTIVE_PLAYER_SEAT:
        assert self._seat_to_player[seat] == player
    for seat, player in self._seat_to_player.items():
      assert self._player_to_seat[player] == seat

    # NOTE: Dealer in terms of *player ID*, not seat number.
    self._dealer = (int(self._dealer) + 1) % self.num_players()

    # Arbitrary choice that we know DEFINITELY represents hitting an infinite
    # loop below if reached.
    definitely_infinite_loop_count = 10000
    # Make sure that we remember to update the above value if someone later
    # drastically increase the max number of players in the game.
    assert definitely_infinite_loop_count > self.num_players() * 10
    loop_counter = 0
    while self._player_to_seat[self._dealer] == INACTIVE_PLAYER_SEAT:
      self._dealer = (int(self._dealer) + 1) % self.num_players()

      loop_counter += 1
      if loop_counter >= definitely_infinite_loop_count:
        raise RuntimeError(
            "Detected infinite loop when attempting to rotate the dealer."
        )

  def update_seat_rotation(self):
    """Rotates seats based on self._dealer.

    Note: deliberately does NOT depend on self._rotate_dealer being True. This
    is because it is entirely valid for a game to not rotate dealers after each
    hand, while still _starting_ with a specific non-default rotation / button
    player that is not the first or last player ID. Similarly, even if we don't
    _normally_ rotate dealer with that setting set to False, in practice we may
    still have to rotate dealer in the event that the dealer busts out.

    Typically called during 1. initialization and 2. immediately after the
    dealer has been updated, following the end of a hand.

    Following this method call, if the game rotates dealers then the seats
    should be updated (to work with pokerkit's seat rotation logic) such that:

    For N > 2 seats: the Button is now the last seat and that the SB is the
    first seat, with players in between proceeding clockwise in standard
    post-flop game order. I.e.
    - SB is seat 0
    - BB (if any) is seat 1
    - UTG (if any) is seat 2
    - ...
    - Button is seat N-1

    For N=2 seats (headsup):
    - the SB / BTN is seat 0
    - the BB is seat 1

    For more details, see the pokerkit docs:
    https://pokerkit.readthedocs.io/en/stable/simulation.html#position
    https://pokerkit.readthedocs.io/en/stable/examples.html
    """
    # At each call of update_seat_assignments_unrotated it effetively 'resets'
    # the rotation entirely, so we cannot simply rotate each seat by 1 forwards.
    # Instead we must compute the exact amount to rotate the player with ID
    # self._dealer such that their seat will end up as the last seat in the
    # list, i.e. _num_active_players - 1.
    #
    # This is 'safe' / requires no modulo because _num_active_players will be
    # strictly greater than the seat number. Since - again - before ever calling
    # this method we would have already reset all the seats / removed any gaps
    # due to players busting out.
    assert self._player_to_seat[self._dealer] >= 0
    assert self.num_players() > self._player_to_seat[self._dealer]
    seat_rotation = (
        self._num_active_players - 1 - self._player_to_seat[self._dealer]
    )
    assert seat_rotation >= 0
    next_player_to_seat = {
        player: (
            (unrotated_seat + seat_rotation) % self._num_active_players
            if unrotated_seat != INACTIVE_PLAYER_SEAT
            else INACTIVE_PLAYER_SEAT
        )
        for player, unrotated_seat in self._player_to_seat.items()
    }
    assert next_player_to_seat[self._dealer] == self._num_active_players - 1
    assert (
        len([
            seat
            for seat in next_player_to_seat.values()
            if seat != INACTIVE_PLAYER_SEAT
        ])
        == self._num_active_players
    )
    self._player_to_seat = next_player_to_seat
    self._seat_to_player = {
        seat: player
        for player, seat in next_player_to_seat.items()
        if seat != INACTIVE_PLAYER_SEAT
    }
    assert all([player >= 0 for player in self._seat_to_player])
    assert len(self._seat_to_player) == self._num_active_players

  def update_blinds_bet_sizes_and_or_bring_in(self):
    """Updates the blinds for the next hand.

    Port of RepeatedPokerState::UpdateBlinds, with a couple of changes to
    accomodate pokerkit's differences from ACPC + also support small_bet,
    big_bet, and bring_in for other poker variants.
    """
    # (No need to update seat assignments for blinds like in
    # RepeatedPoker.UpdateBlinds. Since pokerkit instead directly maps each
    # stack to table position based on index, which we're handling elsewhere
    # via seat rotations.)

    def update_blind_bring_in_or_bet_size(level):
      if isinstance(level, BlindLevel):
        self._small_blind = level.small_blind
        self._big_blind = level.big_blind
      elif isinstance(level, BetSizeLevel):
        self._small_bet_size = level.small_bet_size
        self._big_bet_size = level.big_bet_size
      elif isinstance(level, BringInLevel):
        self._bring_in = level.bring_in
      else:
        raise ValueError(f"Unsupported schedule level type: {type(level)}")

    for schedule in [
        self._blind_schedule_levels,
        self._bring_in_schedule_levels,
        self._bet_size_schedule_levels,
    ]:
      if schedule:
        num_hands = 0
        for level in schedule:
          if self._hand_number < num_hands + level.num_hands:
            update_blind_bring_in_or_bet_size(level)
            break
          num_hands += level.num_hands
        else:
          # If we've exceeded the schedule (and never break-d above), use the
          # last level.
          update_blind_bring_in_or_bet_size(schedule[-1])

  def update_pokerkit_wrapper(self):
    """Updates the underlying PokerkitWrapper and State.

    Port of RepeatedPokerState::UpdateUniversalPoker, with a couple of changes
    to accommodate the differences between pokerkit and ACPC.
    """
    # NOTE: Updating the ._pokerkit_game_params here, NOT the original params
    # stored in the RepeatedPokerkitGame above!
    self._pokerkit_game_params["num_players"] = self._num_active_players
    stacks = [None] * self._num_active_players
    for p in range(self.num_players()):
      if self._player_to_seat[p] != INACTIVE_PLAYER_SEAT:
        # NOTE: Unlike RepeatedPoker we need to ensure that these are actually
        # in the specific 'rotated' seat order (due to how pokerkit works).
        stacks[self._player_to_seat[p]] = self._stacks[p]
    self._pokerkit_game_params["stack_sizes"] = " ".join(str(s) for s in stacks)

    if (
        self._small_blind != INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
        or self._big_blind != INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE
    ):
      self._pokerkit_game_params["blinds"] = (
          f"{self._small_blind} {self._big_blind}"
      )
    else:
      if "blinds" in self._pokerkit_game_params:
        del self._pokerkit_game_params["blinds"]

    # Unlike blinds or stack_sizes, the remaining params are all individually
    # passed in. Allowing us to easily handle them all in the exact same way.
    def set_or_remove_from_params(param, value):
      if value != INVALID_BLIND_BET_SIZE_OR_BRING_IN_VALUE:
        self._pokerkit_game_params[param] = value
      else:
        if param in self._pokerkit_game_params:
          del self._pokerkit_game_params[param]

    set_or_remove_from_params("bring_in", self._bring_in)
    set_or_remove_from_params("small_bet", self._small_bet_size)
    set_or_remove_from_params("big_bet", self._big_bet_size)

    # No need to determine a 'first player' like in
    # RepeatedPoker::UpdateUniversalPoker seat we handle this all via seat
    # rotations.

    # Finally, load a new pokerkit_wrapper game and state with the updated
    # params.
    self.pokerkit_wrapper_game = pyspiel.load_game(
        # NOTE: Using the just-updated ._pokerkit_game_params here, NOT the
        # original params passed into the RepeatedPokerkitGame above!
        pyspiel.game_parameters_to_string(self._pokerkit_game_params)
    )
    self.pokerkit_wrapper_state = self.pokerkit_wrapper_game.new_initial_state()

  def _apply_action(self, action):
    """Port of RepeatedPokerState::DoApplyAction for PokerkitWrapper."""
    self.pokerkit_wrapper_state.apply_action(action)
    if not self.pokerkit_wrapper_state.is_terminal():
      return

    assert not self.pokerkit_wrapper_state._wrapped_state.status

    # Record hand-level information.
    for i, per_seat_returns in enumerate(self.pokerkit_wrapper_state.returns()):
      player: int = self._seat_to_player[i]
      self._hand_returns[-1][player] = per_seat_returns
    wrapped_state = self.pokerkit_wrapper_state._wrapped_state
    self._wrapped_state_hand_histories.append(str(wrapped_state))
    # Terminate or start a new hand
    if self._hand_number + 1 == self.get_game()._max_num_hands:
      self._is_terminal = True
      return
    self._hand_number += 1
    self._hand_returns.append([0.0] * self.num_players())
    self.update_stacks()
    self.update_seat_assignments_unrotated()
    if self._num_active_players == 1:
      # We're playing a tournament and we have a winner
      self._is_terminal = True
      return
    self.update_dealer()
    self.update_seat_rotation()
    self.update_blinds_bet_sizes_and_or_bring_in()
    self.update_pokerkit_wrapper()

  def current_player(self):
    """Returns the current player."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    elif self.is_chance_node():
      return pyspiel.PlayerId.CHANCE
    else:
      assert self.pokerkit_wrapper_state.current_player() not in [
          pyspiel.PlayerId.TERMINAL,
          pyspiel.PlayerId.CHANCE,
      ]
      assert self.pokerkit_wrapper_state.current_player() >= 0
      return self._seat_to_player[self.pokerkit_wrapper_state.current_player()]

  # pylint: disable=g-bad-todo (copied from RepeatedPoker)
  # TODO: jhtschultz - Switch to rewards?
  # pylint: enable=g-bad-todo
  def returns(self):
    """Returns the total returns for each player.

    Port of RepeatedPokerState::Returns.
    """
    assert self._hand_number + 1 == len(self._hand_returns)
    returns = [0.0] * self.num_players()
    if not self.is_terminal():
      return returns
    for hand_returns in self._hand_returns:
      for i in range(self.num_players()):
        returns[i] += hand_returns[i]
    return returns

  def to_string(self):
    """Returns a string representation of the state.

    Port of RepeatedPokerState::ToString.
    """
    return (
        f"Hand {self._hand_number}\n{self.pokerkit_wrapper_state.to_string()}"
    )

  # Deliberately not porting RepeatedPokerState::ObservationString. See
  # make_py_observer + RepeatedPokerkitObserver for the observer implementation.

  def clone(self):
    """Creates a copy of the state."""
    return copy.deepcopy(self)

  def _legal_actions(self, player):
    if self.is_terminal():
      raise ValueError(
          "Attempted to get legal actions for player {player} when the game is"
          " terminal."
      )
    if self.is_chance_node():
      raise ValueError(
          "Attempted to get legal actions for player {player} when the game is"
          " a chance node."
      )
    return self.pokerkit_wrapper_state.legal_actions(
        self._player_to_seat.get(int(player))
    )

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self.is_chance_node()
    assert self.pokerkit_wrapper_state.chance_outcomes()
    return self.pokerkit_wrapper_state.chance_outcomes()

  def _action_to_string(self, player, action):
    if self.is_chance_node():
      if player != pyspiel.PlayerId.CHANCE:
        raise ValueError("Chance actions must be played by the chance player.")
      # Deliberately NOT mapping to seats. Chance nodes don't have a seat and
      # will use the same ID value here + in the pokerkit_wrapper_state at all
      # times (unlike our 'real' players).
      return self.pokerkit_wrapper_state.action_to_string(player, action)
    else:
      if not player >= 0 or player >= self.num_players():
        raise ValueError(
            f"Player {player} is out of range [0, {self.num_players()})"
        )
      return self.pokerkit_wrapper_state.action_to_string(
          self._player_to_seat.get(int(player)), action
      )

  def is_terminal(self):
    return self._is_terminal

  def is_chance_node(self):
    return self.pokerkit_wrapper_state.is_chance_node()

  def __str__(self):
    return f"Hand number: {self._hand_number}\n{self.pokerkit_wrapper_state}"


class RepeatedPokerkitObserver:
  """RepeatedPokerkit "Observer" for creating per-player state tensors and strings."""

  def __init__(self, game, iig_obs_types, params):
    """Initializes the RepeatedPokerkitObserver."""
    if params:
      raise ValueError(
          f"Observation parameters not supported; passed {params}."
      )
    self.game = game
    self.iig_obs_types = iig_obs_types
    if self.iig_obs_types is None:
      # Effectively: default to returning observation_* instead of
      # information_state_* if not specified. Since `perfect_recall` in practice
      # chooses between information_state_* vs observation_* from the
      # perspective of OpenSpiel. See e.g.
      # python/algorithms/generate_playthrough.py
      self.iig_obs_types = pyspiel.IIGObservationType(perfect_recall=False)
    self.params = params

    # Not actually used since `provides_observation_tensor=False`, but necessary
    # since some tests (incorrectly?) assume these always have been created /
    # test for the attribute's existence.
    self.tensor = np.array([])
    self.dict = {}

  def set_from(self, state, player) -> None:
    """No-op; see `provides_observation_tensor=False,` above.

    Defined only because some tests (incorrectly?) assume this method will
    always be defined.

    Args:
      state: The state to extract the observation from.
      player: The player to extract the observation for.
    """
    if player < 0 or player >= state.num_players():
      raise ValueError(
          f"Player {player} is out of range [0, {state.num_players()})"
      )
    pass

  def string_from(self, state, player) -> str:
    """An observation of `state` from the PoV of `player`, as a string.

    If the Observer's iig_obs_types `perfect_recall` is True, then this is
    equivalent to returning an `information_state_string`. If `perfect_recall`
    is False, then this is equivalent to returning an `observation_string`.

    Args:
      state: The state to extract the observation from.
      player: The player to extract the observation for.

    Returns:
      A string representation of the observation.
    """
    if self.iig_obs_types.perfect_recall:
      # Playthrough tests unfortunately do not properly respect the
      # `provides_information_state_string=False` setting in the game
      # definition, so we need to avoid raising an error here.

      # raise ValueError( "Information state string not supported.")
      return "Information state string not supported."

    assert player >= 0
    # pytype: disable=protected-access
    assert player < state.num_players()
    seat_id = state._player_to_seat.get(player)
    # pytype: enable=protected-access
    if seat_id == INACTIVE_PLAYER_SEAT:
      # pylint: disable=g-bad-todo (copied from RepeatedPoker)
      # TODO: jhtschultz - consider adding an observer to pokerkit and returning
      # the public information here. This would allow players who have been
      # eliminated to continue watching the game.
      # pylint: enable=g-bad-todo
      return "Game over.\n"
    pokerkit_wrapper_observer = state.pokerkit_wrapper_game.make_py_observer(
        iig_obs_type=pyspiel.IIGObservationType(perfect_recall=False)
    )
    return pokerkit_wrapper_observer.string_from(
        state.pokerkit_wrapper_state, seat_id
    )


# TODO: b/437724266 - remove once no longer disabling at the top of this file.
# pylint: enable=protected-access

pyspiel.register_game(_GAME_TYPE, RepeatedPokerkit)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
r"""Pokerkit wrapper to provide support for multiple poker game variants.

The base class supports:
- Texas Hold'em
- Limit Texas Hold'em
- Seven Card Stud
- Razz
- Pot Limit Omaha (PLO)

We also provide a subclass that more closely mimics the UniversalPoker / ACPC
implementation in the C++ codebase, but only supports:
- Texas Hold'em
- Limit Texas Hold'em
"""

import copy
import dataclasses
import decimal
import inspect
import math

from absl import logging
import numpy as np
import pokerkit

import pyspiel


# For documentation on using pokerkit see
# https://pokerkit.readthedocs.io/en/0.4/reference.html, and most notably
# https://pokerkit.readthedocs.io/en/0.4/reference.html#module-pokerkit.state

# Automatic steps we want performed by pokerkit without explicit function calls.
# The selection here is based of the example comments in pokerkit/pyspiel.py.
_DEFAULT_AUTOMATIONS = (
    pokerkit.Automation.ANTE_POSTING,
    pokerkit.Automation.BET_COLLECTION,
    pokerkit.Automation.BLIND_OR_STRADDLE_POSTING,
    pokerkit.Automation.HOLE_CARDS_SHOWING_OR_MUCKING,
    pokerkit.Automation.HAND_KILLING,
    pokerkit.Automation.CHIPS_PUSHING,
    pokerkit.Automation.CHIPS_PULLING,
    # WARNING: do NOT automate CARD_BURNING, if for whatever reason you decide
    # to go re-enabling it. (If you want to do card burning, to be faithful to
    # OpenSpiel you'll need to handle it via chance nodes / explicit actions!)
)


# To add support for additional pokerkit game variants add them to this map.
_SUPPORTED_VARIANT_CLASSES = [
    pokerkit.NoLimitTexasHoldem,
    pokerkit.NoLimitShortDeckHoldem,
    # "Limit Texas Hold'em"
    pokerkit.FixedLimitTexasHoldem,
    # "Seven Card Stud"
    pokerkit.FixedLimitSevenCardStud,
    # "Razz"
    pokerkit.FixedLimitRazz,
    # "Pot Limit Omaha (PLO)"
    pokerkit.PotLimitOmahaHoldem,
]
SUPPORTED_VARIANT_MAP = {
    # e.g. { "NoLimitTexasHoldem": pokerkit.NoLimitTexasHoldem, ... }
    variant.__name__: variant
    for variant in _SUPPORTED_VARIANT_CLASSES
}

_SHORT_DECK_VARIANTS = [
    pokerkit.NoLimitShortDeckHoldem.__name__,
]

_STANDARD_DECK_VARIANTS = [
    v.__name__
    for v in _SUPPORTED_VARIANT_CLASSES
    if v.__name__ not in _SHORT_DECK_VARIANTS
]

_VARIANTS_SUPPORTING_ACPC_STYLE = [
    pokerkit.NoLimitTexasHoldem.__name__,
    pokerkit.FixedLimitTexasHoldem.__name__,
]


def _get_variants_supporting_param(param_name: str) -> list[str]:
  return [
      name
      for name, cls in SUPPORTED_VARIANT_MAP.items()
      if param_name in inspect.signature(cls).parameters.keys()
  ]


VARIANT_PARAM_USAGE = {
    param: _get_variants_supporting_param(param)
    for param in [
        "bring_in",
        "small_bet",
        "big_bet",
        "raw_blinds_or_straddles",
        "max_completion_betting_or_raising_count",
    ]
}

# WARNING: do not use comma-containing string param values here. Also, do not
# use strings for single-value integers. Instead respectively use space (" ")
# separation inside your strings values or plain numeric values (e.g. ints).
#
# If you don't, you may run into problems with the default pickle
# serialization/deserialization - e.g. errors or crashes on a round-trip.
# (This would potentially be fixable via custom __getstate__ and __setstate__
# methods, but we would have to define and maintain them ourselves if so.)
_DEFAULT_PARAMS = {
    # Matches pokerkit.NoLimitTexasHoldem.__name__
    # but kept as a hardcoded string to provide clarity to readers.
    # For more details, see the pokerkit docs.
    "variant": "NoLimitTexasHoldem",
    "num_players": 2,
    # Assumed to be "<SB> <BB>", i.e. space-separated numbers (within the
    # string).
    # NOTE: Small Blind must be >= 1, Big Blind must be >= 2, and SB <= BB.
    # TODO: b/437724266 - Add support for any N blinds where 0 < N < num_players
    # instead of requiring exactly "<SB> <BB>". (Pokerkit already supports this,
    # we just need to ensure we're handling things properly on our end.)
    "blinds": "5 10",
    "num_streets": 4,
    # Relevant for games with public hole cards. Will be ignored by all
    # other games (including the default NoLimitTexasHoldem).
    # NOTE: Must be strictly < small_bet.
    "bring_in": 5,
    # Relevant for various FixedLimit ("limit") games. Will be ignored by all
    # other games (including the default NoLimitTexasHoldem). Must be provided
    # if bring_in is provided.
    # NOTE: Must be strictly < big_bet.
    "small_bet": 10,
    "big_bet": 20,
    # NOTE: not yet officially supported. Added primarily for future-proofing.
    "antes": "0 0",
    # Stack sizes for each player. Should be a string containing num_players #
    # of space-separated integers.
    #
    # NOTE: Pokerkit considers "positions", not "seat numbers". As per the docs
    # https://pokerkit.readthedocs.io/en/0.4/simulation.html#position,
    # """
    # For non-heads-up button games, this means that the zeroth player will
    # always be the small blind, the first player will always be the big blind,
    # the second player will always be UTG, and so on until the last player who
    # will always be in the button and in position. The notion of position does
    # not really exist in stud games. In this case, the player to the immediate
    # left of the dealer should be in the zeroth position while the player to
    # the immediate right of the dealer should be in the last position.
    # """
    "stack_sizes": "2000 2000",
}

_GAME_TYPE = pyspiel.GameType(
    short_name="python_pokerkit_wrapper",
    long_name="Python Pokerkit",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=10,  # Arbitrarily chosen to match universal_poker
    min_num_players=2,  # Arbitrarily chosen to match universal_poker
    provides_information_state_string=True,
    provides_observation_string=True,
    # Tensors are not yet supported. Instead use the corresponding strings.
    provides_information_state_tensor=False,
    provides_observation_tensor=False,
    # TODO: b/437724266 - determine what this value does and whether it should
    # be True or False / what functionality it's actually controlling.
    provides_factored_observation_string=False,
    parameter_specification=_DEFAULT_PARAMS,
)

# Used in the PokerkitWrapperAcpcStyle subclass; see below for more details.
_GAME_TYPE_ACPC_STYLE = pyspiel.GameType(
    short_name="python_pokerkit_wrapper_acpc_style",
    long_name="Python Pokerkit ACPC Style (mimicking UniversalPoker / ACPC)",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=10,  # Arbitrarily chosen to match universal_poker
    min_num_players=2,  # Arbitrarily chosen to match universal_poker
    # TODO: b/434776281 - Toggle True again once *properly* supported via the
    # Observer class when its iig_obs_type.perfect_recall is True.
    provides_information_state_string=False,
    provides_information_state_tensor=False,
    provides_observation_tensor=True,
    provides_observation_string=True,
    # TODO: b/437724266 - determine what this value does and whether it should
    # be True or False / what functionality it's actually controlling.
    provides_factored_observation_string=False,
    parameter_specification=_DEFAULT_PARAMS,
)


# NOTE: FOLD and CHECK_OR_CALL deliberately chosen to match kFold and kCall in
# universal_poker so that it's easier to port code over from it.
ACTION_FOLD = 0
ACTION_CHECK_OR_CALL = 1
FOLD_AND_CHECK_OR_CALL_ACTIONS = [ACTION_FOLD, ACTION_CHECK_OR_CALL]

# NOTE: with one exception (see below), all player actions with value >= 2 will
# be interpreted as part of a raise, reraise, or completion bet. Generally this
# follows typical poker assumptions, where "raise to X" declares a total
# contribution on the *current street* of X chips. (This differs from an
# ACPC-"style" implementation / universal_poker's approach, which isntead treats
# it as declaring a total contribution of X chips across **all** streets", i.e.
# the entire hand).

# NOTE: As a result of using action `1` to handle ACTION_CHECK_OR_CALL, we have
# to handle bet sizes of 1 chip via special-case 'mapping' them to another
# action value. This can can happen in exactly one case: when a player has a
# single chip left in their stack and wants to shove all-in. (We prevent bet
# sizes < 2 in all other situations by ensuring that we only ever provide
# pokerkit min_bet and small_bet parameters of size >= 2.)


def _parse_values(param_str) -> list[int | decimal.Decimal]:
  """Convert " "-separated values in a ``str`` to numerical values via pokerkit.

  Args:
    param_str: A space-separated string of numerical values.

  Returns:
    A list of numerical values, which can be either `int` or `decimal.Decimal`.
  """
  if not param_str:
    return None
  return [pokerkit.parse_value(s) for s in param_str.split(" ")]


# ------------------------------------------------------------------------------
# --- "Base-class" PokerkitWrapper implementation. ---
#


# (Note: Subclassed below by PokerkitWrapperAcpcStyle)
class PokerkitWrapper(pyspiel.Game):
  """An OpenSpiel game that wraps pokerkit."""

  def __init__(self, params=None):
    game_type = self._game_type()

    if not params:
      params = {}
    # Replicates pyspiel.load_game()'s behavior of using the default's value for
    # each param missing from the provided params.
    self.params = _DEFAULT_PARAMS | params
    # Prevents accidental usage params instead of self.params below.
    # NOTE: Only removes the reference from *local scope* here; for details see:
    # https://docs.python.org/3/tutorial/classes.html#python-scopes-and-namespaces
    del params

    variant = self.params.get("variant")
    if variant not in SUPPORTED_VARIANT_MAP:
      raise ValueError(f"Unknown poker variant: {variant}")

    # **** **** **** **** **** **** **** **** **** **** **** **** **** ****
    # WARNING: DO NOT CALL self.num_players() BELOW UNTIL AFTER COMPLETELY
    # FINISHING INITIALIZATION OF THIS OBJECT.
    #
    # Instead, for now only use this `num_players` local variable.
    # **** **** **** **** **** **** **** **** **** **** **** **** **** ****
    #
    # Context: Calling self.num_players() too early may result in segfaults.
    # Which have historically been very challenging + time-consuming to debug
    # due to lacking clear error messages.
    #
    # (Please also do not remove this comment; LLMs love quietly slipping in
    # usage of self.num_players or self.num_players(). This giant warning is
    # helpful for reminding them that doing so is a bad idea.)
    num_players = self.params["num_players"]
    if num_players > game_type.max_num_players:
      raise ValueError(
          f"Players in game ({num_players}) exceeds"
          f" max_players ({game_type.max_num_players})."
      )
    if num_players < game_type.min_num_players:
      raise ValueError(
          f"Players in game ({num_players}) is less than"
          f" min_players ({game_type.min_num_players})."
      )

    # NOTE: Will not include any params set by a user which happen to match the
    # default value. Meaning even params not present in this set might have been
    # set explicitly by a user; we cannot assume it came from a default value.
    definitely_overriden = set([
        param
        for param, v in _DEFAULT_PARAMS.items()
        if v != self.params.get(param)
    ])

    # --- Blinds, Bet Sizes, and Antes or Bring-ins ---
    self.blinds = _parse_values(self.params["blinds"])
    if (
        self.blinds is None
        or len(self.blinds) != 2
        or self.blinds[0] > self.blinds[1]
    ):
      raise ValueError(
          "blinds must parse to a list of 2 values [Small Blind, Big Blind],"
          " where SB >= 1, BB >= 2, and SB <= BB. But instead got: ",
          self.params["blinds"],
          " which parsed to: ",
          self.blinds,
      )
    if any(x <= 0 for x in self.blinds):
      raise ValueError("Blinds must all be >= 1: ", self.blinds)
    # Ensure that users do not configure games with blinds low enough that it
    # will matter that we can't support bet sizes <= 1.
    if not max(self.blinds) >= 2:
      raise ValueError("Big Blind must be at least 2: ", self.blinds)

    assert self.params["small_bet"] < self.params["big_bet"]
    if variant in VARIANT_PARAM_USAGE["bring_in"]:
      assert self.params["bring_in"] < self.params["small_bet"]

    # This is such a common + confusing mistake that it's helpful to at least
    # warn about it, even if we can't actually throw an error (without losing
    # the ability to manually choose these default values in certain cases).
    for param1, param2 in [
        ("bring_in", "small_bet"),
        ("bring_in", "big_bet"),
        ("small_bet", "big_bet"),
        ("big_bet", "small_bet"),
    ]:
      if (
          param1 in definitely_overriden
          and param2 not in definitely_overriden
          and (
              variant in VARIANT_PARAM_USAGE[param1]
              or variant in VARIANT_PARAM_USAGE[param2]
          )
      ):
        logging.warning(
            "In the following pair of tightly-coupled params, the first param"
            " differs from its default value but the second does not: %s vs"
            " %s. (NOTE: This may be intended, and is technically allowed, so"
            " we are not raising an error. But in practice, it's very likely"
            " that you set the params here incorrectly!).",
            param1,
            param2,
        )

    stack_sizes = _parse_values(self.params["stack_sizes"])
    if stack_sizes is None or len(stack_sizes) != num_players:
      raise ValueError(
          "stack_sizes must be a space-separated list of "
          f" {num_players} values. Was provided: {self.params['stack_sizes']},"
          f" which parsed to: {stack_sizes}"
      )
    self.stack_sizes = stack_sizes

    # --- Card and Deck Information ---
    self._deck = pokerkit.Deck.STANDARD
    if variant in _SHORT_DECK_VARIANTS:
      self._deck = pokerkit.Deck.SHORT_DECK_HOLDEM
    else:
      assert variant in _STANDARD_DECK_VARIANTS

    self.deck_size = len(self._deck)
    self.card_to_int = {card: i for i, card in enumerate(self._deck)}
    self.int_to_card = {i: card for i, card in enumerate(self._deck)}
    self.num_streets = self.params.get("num_streets")

    # --- GameInfo and GameType Setup ---

    # Initialize to None since the function is going to attempt to check this
    # to see if we cached it already.
    self._max_game_length_estimate = None
    self._max_game_length_estimate = self._calculate_max_game_length()

    self._game_info = pyspiel.GameInfo(
        num_distinct_actions=self._calculate_num_distinct_actions(),
        max_chance_outcomes=self.deck_size,
        num_players=num_players,
        min_utility=self._calculate_min_utility(),
        max_utility=self._calculate_max_utility(),
        utility_sum=0.0,
        max_game_length=self._calculate_max_game_length(),
    )
    super().__init__(game_type, self._game_info, self.params)

  def _calculate_num_distinct_actions(self):
    """Returns an upper bound on the number of distinct player actions.

    NOTE: This could be a bit tighter, e.g. depending on the value of min_bet
    or specifics of the game variant. That said, this just needs to be an upper
    bound, so it should be entirely fine.
    """
    # +1 is to account for ACTION_FOLD being 0
    return max(self.stack_sizes) + 1

  def _calculate_min_utility(self):
    """Returns an upper-bound (magnitude) on minimum utility for the game.

    NOTE: for sake of simplicity, this base class utility estimate assumes that
    in all games the deepest-stack will always be able to get all the chips in
    AND actually lose them. This is not as tight as it could be in many cases,
    e.g. where effective stack sizes or bet limits prevent the deepest-stack
    from actually doing so. If you need tighter bounds, consider subclassing and
    overriding.
    """
    # TODO: b/437724266 - use effective stack size instead of raw stack size.
    return float(-max(self.stack_sizes))

  def _calculate_max_utility(self):
    """Returns an upper-bound on maximum utility for the game.

    NOTE: for sake of simplicity, this base class utility estimate assumes that
    in all games the deepest-stack will always be able to win all other players'
    chips. This is not as tight as it could be in many cases, e.g. where
    bet limits prevent the deepest-stack from actually doing so. If you need
    tighter bounds, consider subclassing and overriding.
    """
    return float(sum(self.stack_sizes) - max(self.stack_sizes))

  def game_info(self):
    return self._game_info

  def _calculate_max_game_length(self):
    """Provides a (very rough) upper bound on the maximum game length."""
    if self._max_game_length_estimate is not None:
      return self._max_game_length_estimate

    # Both theoretically and in practice, games will be MUCH shorter than this.
    # However, this does set an ironclad upper bounds on the game length,
    # which is 'good enough' for our purposes.
    # NOTE: Assumes that in nolimit games, min_bet is exactly one big blind.
    min_reraise_amount = max(self.blinds)
    return (
        # Upper bound on check actions from players not betting, on all streets.
        self.params.get("num_players") * max(self.params.get("num_streets"), 1)
        # Upper bound on reraises + calls by the other players. (This is a
        # factor of total chips, and so doesn't need to consider streets.)
        + (max(self.stack_sizes) // min_reraise_amount)
        * max(1, self.params.get("num_players") - 1)
        # Upper bound on performing all deal actions
        + self.deck_size
    )

  def new_initial_state(self):
    return PokerkitWrapperState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    return PokerkitWrapperObserver(self, iig_obs_type, params)

  def raise_error_if_player_out_of_range(self, player):
    if player < 0 or player >= self.num_players():
      raise ValueError(
          f"Player {player} is out of range [0, {self.num_players()})"
      )

  # Not yet actually supported. Providing this 'return hardcoded value'
  # placeholder implementation instead of throwing an error to avoid crashing
  # anything in OpenSpiel that doesn't properly respect the
  # provides_information_state_tensor=False setting above.
  def information_state_tensor_size(self) -> int:
    logging.warning(
        "information_state_tensor_size() is not yet supported for"
        " PokerkitWrapper. Returning a default value of 1 for now."
    )
    return 1

  def _wrapped_game_template_factory(self):
    """Creates a default pokerkit 'template' that can easily construct states.

    For more details, see:
    https://pokerkit.readthedocs.io/en/stable/simulation.html#pre-defined-games
    ```
    In certain use cases, one may want to create a template from which just the
    starting stacks and the number of players can be specified. In PokerKit,
    this can be done by creating an instance of the game that acts as a state
    factory from which states are initialized.
    ```

    NOTE: By definition, this CANNOT disable card burning, as it's using the
    default pokerkit game variants (all of which have card burning enabled /
    do not allow custom Street inputs). If using this to create a
    wrapped_state you'll want to pass in a custom Streets to disable it!

    WARNING: Don't attempt to merge this into wrapped_state_factory(). You'll
    need this for multiple reasons, including those other than creating a game
    object to track game state. Notably, we have to pass the Pokerkit.Poker into
    calls to HandHistory.from_game_state(...) to construct PHHs.

    Returns:
      A "variant" sub-class pokerkit.Poker object.
    """
    # TODO: b/437724266 - Stop directly accessing values from the params and
    # instead grab the values we need via getters.
    params = self.params
    variant = params.get("variant")
    poker_variant_class = SUPPORTED_VARIANT_MAP.get(variant)
    if not poker_variant_class:
      raise ValueError(f"Unknown / unsupported poker variant: {variant}")

    def in_scope_for_variant(k):
      return k in inspect.signature(poker_variant_class).parameters.keys()

    assert _parse_values(params.get("stack_sizes")) == self.stack_sizes
    game_args = {
        k: v
        for k, v in {
            "automations": _DEFAULT_AUTOMATIONS,
            "ante_trimming_status": False,
            "player_count": params.get("num_players"),
            # As per pokerkit docs:
            # """
            # In Pokerkit, the term ``raw`` is used to denote the fact that
            # they can be supplied in many forms and will be "parsed" or
            # "evaluated" further to convert them into a more ideal form.
            # For instance, ``0`` will be interpreted as no ante for all
            # players. Another value will be interpreted as that value as the
            # antes for all. ``[0, 2]`` and ``{1: 2} will be considered as the
            # big blind ante whereas ``{-1: 2}`` will be considered as the
            # button ante.
            # """
            "raw_antes": _parse_values(params.get("antes")),
            "raw_starting_stacks": _parse_values(params.get("stack_sizes")),
            "raw_blinds_or_straddles": self.blinds,
            # Unlike most other args, we don't support custom user inputs for
            # this. Despite the name, it's only used in nolimit games (NOT any
            # of the fixed limit variants, which compute their minimum bet size
            # in other ways). And in practice, almost everyone just uses the big
            # blind as the minimum bet size in nolimit games.
            "min_bet": max(self.blinds),  # Always set to the big blind
            # All remaining args we need line up exactly with the params
            # key/values.
        }.items()
        if in_scope_for_variant(k)
    } | {k: v for k, v in params.items() if in_scope_for_variant(k)}

    # Double check that we didn't somehow end up in a situation where we are
    # allowing bet sizes that could conflict with the reserved values for our
    # fold / check actions.
    min_bet = game_args.get("min_bet")
    if min_bet is not None:
      assert min_bet > 0
      assert min_bet not in FOLD_AND_CHECK_OR_CALL_ACTIONS
    # TODO: b/434776281 - consider moving additional validation checks out of
    # the constructor and into this spot here right before we actually use them.

    return poker_variant_class(**game_args)

  def wrapped_state_factory(self) -> pokerkit.State:
    """Creates the "wrapped" (not "PokerkitWrapper...") pokerkit.state object.

    As a convienence, this method additionally disables card burning on all
    streets. (Card burning would have no purpose here since our cards' backs
    cannot be marked, unlike with e.g. real physical poker cards).

    Returns:
      A `pokerkit.State` object configured with the game parameters.
    """
    # For more details on why this works (i.e. only two arguments), see
    # the docstring in the _wrapped_game_template_factory() method.
    #
    # NOTE: All default pokerkit games have card burning enabled. **This
    # cannot be prevented, whether via create_state OR via the 'template'
    # approach we use here.** We'll need to create our own state object
    # directly using the pokerkit.State constructor to (properly) disable it.
    game_template = self._wrapped_game_template_factory()
    stacks = _parse_values(self.params.get("stack_sizes"))
    number_players = self.params.get("num_players")
    helper_state = game_template(stacks, number_players)

    # Doing this instead of just directly mutating state.streets as per
    # https://pokerkit.readthedocs.io/en/stable/tips.html#read-only-and-read-write-values
    # """
    #  [The attributes of pokerkit.state.State... should never be modified...
    #  [I]nstead let PokerKit modify them through public method calls. In other
    #  words, the user must only read from the state’s attributes or call public
    #  methods (which may modify them).
    # """
    streets_override = [
        pokerkit.state.Street(
            **dataclasses.asdict(street)
            | {
                "card_burning_status": False,
            }
        )
        for street in helper_state.streets
    ]
    args = copy.deepcopy(
        {
            # A select few args have "raw_" prefixes in the State constructor.
            # Unfortunately, although the constructed State objects have both
            # versions, in practice the non-"raw_" versions are the only ones
            # that are actually (reliably) set to non-None values.
            # As such we have to source these "raw_" args from the equivalent
            # non-"raw_" versions.
            #
            # For example:
            # - giving `raw_starting_stacks` arg `state.starting_stacks`'s value
            # - giving `raw_antes` arg `state.antes`'s value
            # - giving `raw_blinds_or_straddles` arg `state.blinds`'s value
            arg: helper_state.__dict__.get(arg.replace("raw_", ""))
            for arg in inspect.signature(pokerkit.State).parameters.keys()
        }
        | {"streets": streets_override}
    )
    assert len(args) == len(inspect.signature(pokerkit.State).parameters.keys())
    assert None not in args.values()
    # NOTE: CANNOT USE THE "template" APPROACH, NOR .create_state() HERE
    # (without allowing card burning - since neither of these options accept an
    # input 'streets').
    returned_state: pokerkit.State = pokerkit.State(**args)
    # Double checks that we didn't make mistakes when disabing card burning just
    # now. (This is a bit overkill, but we REALLY don't want there to be any
    # chance of accidentally leaving card burning enabled anywhere).
    for street in returned_state.streets:
      assert not street.card_burning_status
    return returned_state

  def _game_type(self):
    return _GAME_TYPE


class PokerkitWrapperState(pyspiel.State):
  """Represents an _OpenSpiel_ 'state' for the PokerkitWrapper game.

  As described by the name, this class indeed wraps a `pokerkit.State` object
  and provides the necessary interface for OpenSpiel's `pyspiel.State`.
  """

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._wrapped_state: pokerkit.State = game.wrapped_state_factory()

  def current_player(self):
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    elif self.is_chance_node():
      return pyspiel.PlayerId.CHANCE
    else:
      return self._wrapped_state.actor_index

  def _legal_actions(self, player):
    return self._legal_actions_base(player)

  def _legal_actions_base(self, player):
    wrapped_state: pokerkit.State = self._wrapped_state
    game: PokerkitWrapper = self.get_game()
    game.raise_error_if_player_out_of_range(player)

    if self.is_terminal():
      raise ValueError(
          f"Attempted to get legal actions for player {player} when the"
          " game is terminal."
      )

    # player is not the active player, and so can't take any actions.
    if player != wrapped_state.actor_index:
      raise ValueError(
          f"Attempted to get legal actions for player {player} when the current"
          f" player is {wrapped_state.actor_index}."
      )

    if not (
        wrapped_state.can_check_or_call()
        or wrapped_state.can_fold()
        or wrapped_state.can_complete_bet_or_raise_to()
    ):
      raise ValueError(
          f"Attempted to get legal actions for player {player} when the"
          " player is not in a position to check, fold, or bet/raise."
      )

    actions = []

    # Handling player nodes first since they're the most straightforward.
    if wrapped_state.can_fold():
      actions.append(ACTION_FOLD)
    if wrapped_state.can_check_or_call():
      actions.append(ACTION_CHECK_OR_CALL)
    if wrapped_state.can_complete_bet_or_raise_to():
      valid_bet_sizes = []
      betting_structure = wrapped_state.betting_structure
      min_bet = wrapped_state.min_completion_betting_or_raising_to_amount
      max_bet = wrapped_state.max_completion_betting_or_raising_to_amount
      assert min_bet is not None and max_bet is not None

      if betting_structure == pokerkit.state.BettingStructure.FIXED_LIMIT:
        # min and max completion, bet, and raise amounts are identical as per:
        # https://pokerkit.readthedocs.io/en/stable/reference.html#pokerkit.state.BettingStructure.FIXED_LIMIT
        assert min_bet == max_bet
        fixed_size: int | None = min_bet
        if (
            fixed_size is not None
            and fixed_size not in FOLD_AND_CHECK_OR_CALL_ACTIONS
        ):
          valid_bet_sizes.append(fixed_size)
      elif (
          betting_structure == pokerkit.state.BettingStructure.NO_LIMIT
          or betting_structure == pokerkit.state.BettingStructure.POT_LIMIT
      ):
        assert min_bet <= max_bet
        valid_bet_sizes.extend([
            amount
            for amount in range(
                min_bet,
                max_bet + 1,
            )
            if amount not in FOLD_AND_CHECK_OR_CALL_ACTIONS
        ])
      else:
        raise RuntimeError(
            f"Unsupported pokerkit betting structure: {betting_structure}"
        )
      assert all(
          amount not in FOLD_AND_CHECK_OR_CALL_ACTIONS
          for amount in valid_bet_sizes
      )

      if len(valid_bet_sizes) == 1:
        only_valid_bet = valid_bet_sizes[0]
        assert only_valid_bet == min_bet and only_valid_bet == max_bet
        current_street: pokerkit.Street = wrapped_state.streets[
            wrapped_state.street_index
        ]
        # "Global minimum" size for the current street, regardless of player.
        # TODO: b/437724266 - double check this actually is independent from
        # player!
        street_minimum_size = (
            current_street.min_completion_betting_or_raising_amount
        )
        if only_valid_bet < street_minimum_size:
          # In 2-person games, the only way for this to happen is if the
          # effective stack size is less than the bet size and the player is
          # (effectively) shoving. So we can trivially verify that this isn't
          # happening incorrectly.
          #
          # NOTE: This cares about _effective_ stack size, not the player's
          # actual stack size! Since this behavior also happens when the *other*
          # player's stack size is less than the minimum for the street.
          if game.num_players() == 2:
            if only_valid_bet != wrapped_state.get_effective_stack(player):
              raise RuntimeError(
                  f"Player {player} has effective stack size of"
                  f" {wrapped_state.get_effective_stack(player)} (and has"
                  f" 'real' stack size of {wrapped_state.stacks[player]} chips)"
                  f"  but only valid bet is {only_valid_bet}, which is"
                  " unexpectedly less than the minimum value for the current"
                  " street AND also doesn't equal their effective stack size."
                  f" {street_minimum_size}."
              )
          # In 3+ person games, this is significantly more complicated to verify
          # since side-pots can significantly complicate the math, especially
          # in fixed-limit games. Any asserts we could write here would be
          # overly bug-prone + require significant testing anyways.
          pass

      # Handle edge case in both no limit and fixed limit where the player has
      # so few chips that a legal bet would have conflicted with one of our
      # reserved actions, and so got filtered out above.
      if not valid_bet_sizes and wrapped_state.stacks[player] > 0:
        # If the stack size is non-zero, then we *should* be in a situation
        # where we were going to try to bet 1 chip, which got filtered out due
        # to it being in the reserved actions set. (Anything else would mean
        # there's a bug in our code above.)
        assert wrapped_state.stacks[player] == 1
        assert min_bet == 1 and max_bet == 1
        # This decision to map to 2 is arbitrary! We could alternatively have
        # mapped to any other (reasonably small) positive integer not in the
        # reserved actions set.
        logging.warning(
            "Mapping shove size 1 chip to action 2 in legal_actions to avoid"
            " conflict with 'reserved' actions.",
        )
        valid_bet_sizes.append(2)

      # TODO: b/437724266 - Extract this entire giant if-statement block out
      # to a helper function to make this overall more readable. (In particular,
      # having the .extend() for the entire can_complete_bet_or_raise_to()
      # handling at the very bottom down here is difficult to read + bug
      # prone).
      assert set(valid_bet_sizes).isdisjoint(FOLD_AND_CHECK_OR_CALL_ACTIONS)
      actions.extend(valid_bet_sizes)

    assert actions
    return sorted(actions)

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities.

    NOTE: Pokerkit has a known-issue of returning certain mucked cards in N>2
    player games. As such, we additionally ourselves 'double check' to ensure
    that there are no cards being incorrectly returned by pokerkit itself. And
    in such cases either A. remove such offending cards, or B. raise an error
    (respectively depending on whether it's this exact 'mucked card' known issue
    vs something else).)

    Raises:
      RuntimeError: If `pokerkit.State.get_dealable_cards()` returns cards that
        are determined to be non-dealable, even after resolving a known issue
        with mucked cards.
    """
    assert self.is_chance_node()
    wrapped_state: pokerkit.State = self._wrapped_state
    # pokerkit has a bug in get_dealable_cards() where it sometimes returns
    # cards in N>2 player games that were already dealt to a different player if
    # that player mucked. We want to filter those out here to avoid bugs
    # in our code later when applying the action (at which point pokerkit will
    # correctly raise an error sayign that the card is not recommened to be
    # dealt).
    # Flattened list of hole cards
    flattened_hole_cards = []
    for sublist in wrapped_state.hole_cards:
      flattened_hole_cards.extend(sublist)
    flattened_board_cards = []
    for sublist in wrapped_state.board_cards:
      flattened_board_cards.extend(sublist)
    double_checked_used_cards = set(wrapped_state.deck) - set(
        flattened_hole_cards
        + flattened_board_cards
        + wrapped_state.burn_cards
        + wrapped_state.mucked_cards
    )
    pokerkit_computed_dealable_cards = set(wrapped_state.get_dealable_cards())
    if pokerkit_computed_dealable_cards != double_checked_used_cards:
      logging.warning(
          "pokerkit.State.get_dealable_cards() did not return the same set of"
          " cards as the set of used cards. %s != %s",
          pokerkit_computed_dealable_cards,
          double_checked_used_cards,
      )
    # Make sure that there's no bugs where *our* logic forgets to filter out
    # cards vs pokerkit's logic
    assert double_checked_used_cards.issubset(pokerkit_computed_dealable_cards)

    for card in pokerkit_computed_dealable_cards - double_checked_used_cards:
      if card in wrapped_state.mucked_cards:
        logging.warning(
            "Confirmed incorrectly-dealable card is an instance of known issue"
            " in pokerkit (card %s is one of the mucked_card). ***REMOVING "
            "CARD %s FROM THE RETURNED CARDS AS A WORKAROUD.****",
            card,
            card,
        )
        pokerkit_computed_dealable_cards.remove(card)
    if pokerkit_computed_dealable_cards != double_checked_used_cards:
      raise RuntimeError(
          "pokerkit.State.get_dealable_cards() still contains cards that we "
          " have determined should not be dealable - even after removing some "
          " cards that were included due to known issues with pokerkit. This"
          " means there are likely MORE problems in pokerkit that we have no"
          " identified yet! Post-'surgery' get_dealable_cards() was"
          f" {pokerkit_computed_dealable_cards}, but when double-checking"
          " ourselves we determined the list should instead be"
          f" {double_checked_used_cards}"
      )

    outcomes = sorted([
        self.get_game().card_to_int[c] for c in pokerkit_computed_dealable_cards
    ])
    p = 1.0 / len(outcomes)
    return [(o, p) for o in outcomes]

  def _apply_action(self, action):
    wrapped_state: pokerkit.State = self._wrapped_state
    # Handling player actions first since it's the more striaghtforward case.
    if not self.is_chance_node():
      # NOTE: using the _base version of the method directly. Since even in
      # situations where legal_actions is overridden, we still want to avoid
      # using the overriding class's actions (which could be different in
      # situations where we're 'mapping' actions in some way).
      if action not in self._legal_actions_base(self.current_player()):
        raise ValueError(
            f"Attempted to apply illegal action {action} to player"
            f" {self.current_player()}. Legal actions were"
            f" {self.legal_actions()}"
        )

      if action == ACTION_FOLD:
        wrapped_state.fold()
      elif action == ACTION_CHECK_OR_CALL:
        wrapped_state.check_or_call()
      else:
        if wrapped_state.can_complete_bet_or_raise_to(action):
          wrapped_state.complete_bet_or_raise_to(action)
        else:
          # Should be an edge case when trying to shove with exactly 1 chip,
          # where we remapped the shove to action 2 to avoid conflicting with
          # reserved actions.
          assert action == 2
          assert wrapped_state.stacks[wrapped_state.actor_index] == 1
          assert wrapped_state.can_complete_bet_or_raise_to(1)
          assert wrapped_state.stacks[wrapped_state.actor_index] == 1
          assert "Bet/Raise to 1 [ALL-IN EDGECASE]" in self.action_to_string(
              wrapped_state.actor_index, action
          )
          wrapped_state.complete_bet_or_raise_to(1)
      return

    # else it's a chance node (which is a bit more complicated):

    card_from_action = self.get_game().int_to_card[action]
    if card_from_action not in wrapped_state.get_dealable_cards():
      raise ValueError(
          f"Chance action maps to non-dealable card {card_from_action}: "
          f"{wrapped_state.can_deal_hole()} "
          f"{wrapped_state.can_deal_board()}"
      )
    # If these ever fail it means that wrapped_state.get_dealable_cards()
    # gave us a card that it shouldn't have!
    # assert card_from_action not in wrapped_state.mucked_cards
    assert card_from_action not in wrapped_state.burn_cards
    assert card_from_action not in wrapped_state.board_cards
    for player_hole_cards in wrapped_state.hole_cards:
      assert card_from_action not in player_hole_cards

    if wrapped_state.can_deal_hole() and wrapped_state.can_deal_board():
      raise ValueError(
          "Chance node with both dealable hole and board cards: "
          f"{wrapped_state.can_deal_hole()} "
          f"{wrapped_state.can_deal_board()}"
      )
    elif wrapped_state.can_deal_hole():
      wrapped_state.deal_hole(card_from_action)
    elif wrapped_state.can_deal_board():
      wrapped_state.deal_board(card_from_action)
    else:
      raise ValueError(
          "Chance node has dealable card, but cannot deal hole or board cards :"
          f"{wrapped_state.can_deal_hole()} "
          f"{wrapped_state.can_deal_board()}"
      )

  def _action_to_string(self, player, action):
    wrapped_state: pokerkit.State = self._wrapped_state
    if self.is_chance_node():
      assert player is not None
      assert player < 0
      if wrapped_state.can_deal_hole():
        return f"Deal Hole Card: {self.get_game().int_to_card[action]}"
      elif wrapped_state.can_deal_board():
        return f"Deal Board Card: {self.get_game().int_to_card[action]}"
      else:
        raise ValueError(
            f"Chance node with action {action} but cannot deal cards. "
            f"can_deal_hole: {wrapped_state.can_deal_hole()} "
            f"can_deal_board: {wrapped_state.can_deal_board()}"
        )
    if player != wrapped_state.actor_index:
      raise ValueError(
          "Attempted to call _action_to_string() for a not-currently-active"
          "player (who shouldn't have any actions available)."
      )
    assert player == wrapped_state.actor_index
    if action == ACTION_FOLD:
      return f"Player {player}: Fold"
    elif action == ACTION_CHECK_OR_CALL:
      amount = wrapped_state.checking_or_calling_amount
      return f"Player {player}: Check" if amount == 0 else f"Call({amount})"
    elif action >= 0 and action not in FOLD_AND_CHECK_OR_CALL_ACTIONS:
      if wrapped_state.stacks[player] != 1:
        return f"Player {player}: Bet/Raise to {action}"

      # In this very specific edge case, we likely mapped a shove size of 1 chip
      # to action 2 to avoid conflicting with reserved actions. For more details
      # see the comment in _legal_actions().
      assert action == 2
      assert wrapped_state.min_completion_betting_or_raising_to_amount == 1
      assert wrapped_state.max_completion_betting_or_raising_to_amount == 1
      # As per the docs:
      # https://pokerkit.readthedocs.io/en/stable/reference.html#pokerkit.state.State.payoffs
      # "Negated versions of these values can be thought of as contributions
      # to the hand by the individual players."
      #
      # Bets on the other hand has the positive amounts the player already
      # contributed on this street.
      contribution_on_prior_streets = -1 * (
          wrapped_state.payoffs[player] + wrapped_state.bets[player]
      )
      assert (
          wrapped_state.starting_stacks[player] - contribution_on_prior_streets
          == 1
      )
      return f"Player {player}: Bet/Raise to 1 [ALL-IN EDGECASE]"
    else:
      raise ValueError(
          f"Invalid action {action} for player {player}. 'legal actions' was "
          f" {self.legal_actions()}"
      )

  def is_terminal(self):
    # .status is True if the state is not terminal, as per
    # https://pokerkit.readthedocs.io/en/stable/reference.html#pokerkit.state.State.status,
    assert self._wrapped_state.status is not None
    is_terminal = not self._wrapped_state.status
    if is_terminal:
      assert isinstance(
          self._wrapped_state.operations[-1], pokerkit.ChipsPulling
      )
    return is_terminal

  def is_chance_node(self):
    # Chance nodes should only refer to situations where a card is being dealt
    # from the deck. (Technically we could do the same for burned cards, but
    # we're automating that phase entirely, so it would regardless always be
    # false).
    return (
        self._wrapped_state.can_deal_hole()
        or self._wrapped_state.can_deal_board()
    )

  def returns(self):
    if not self.is_terminal():
      return [0.0 for _ in range(self.get_game().num_players())]
    payoffs = [float(p) for p in self._wrapped_state.payoffs]

    # Double check that this is consistent with the rest of pokerkit.State's
    # internal stack and starting_stack values to detect any bugs.
    for player, payoff in enumerate(payoffs):
      assert payoff == (
          self._wrapped_state.stacks[player]
          - self._wrapped_state.starting_stacks[player]
      )

    return payoffs

  def clone(self):
    return copy.deepcopy(self)

  def deepcopy_wrapped_state(self) -> pokerkit.State:
    """Create and return a deepcopy of the 'wrapped' pokerkit.state."""
    return copy.deepcopy(self._wrapped_state)

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    state: pokerkit.State = self._wrapped_state
    parts = []
    parts.append(f"Stacks: {state.stacks}")
    parts.append(f"Bets: {state.bets}")
    parts.append(f"Board: {state.board_cards}")
    # WARNING: Will include private hole cards for all players!
    parts.append(f"Hole Cards: {state.hole_cards}")
    parts.append(f"Pots: {list(state.pots)}")
    # WARNING: Will include private hole cards for all players!
    parts.append(f"Operations: {state.operations}")
    return " | ".join(parts)


class PokerkitWrapperObserver:
  """Pokerkit "Observer" for creating per-player state tensors and strings."""

  def __init__(self, game, iig_obs_types, params):
    """Initializes the PokerkitWrapperObserver."""
    if params:
      raise ValueError(
          f"Observation parameters not supported; passed {params}."
      )
    self.game = game
    self.iig_obs_types = iig_obs_types
    if self.iig_obs_types is None:
      # Effectively: default to returning observation_* instead of
      # information_state_* if not specified. Since `perfect_recall` in practice
      # chooses between information_state_* vs observation_* from the
      # perspective of OpenSpiel. See e.g.
      # open_spiel/python/algorithms/generate_playthrough.py
      self.iig_obs_types = pyspiel.IIGObservationType(perfect_recall=False)
    self.params = params

    # Not actually used since `provides_observation_tensor=False`, but necessary
    # since some tests (incorrectly?) assume these always have been created /
    # test for the attribute's existence.
    self.tensor = np.array([])
    self.dict = {}

  def _poker_hand_history_actions(self, state, player) -> list[str]:
    """Returns PHH (Poker Hand History) actions from player's perspective."""
    game = state.get_game()
    game.raise_error_if_player_out_of_range(player)

    # For context see
    # https://pokerkit.readthedocs.io/en/stable/notation.html#writing-hands
    # and https://phh.readthedocs.io/en/stable/required.html
    hand_history = pokerkit.HandHistory.from_game_state(
        # pylint: disable=protected-access
        # NOTE: Technically has card burning enabled! But should be close enough
        # for our purposes here - in practice still results in the same string.
        game._wrapped_game_template_factory(),
        # Used as an alternative to .deepcopy_wrapped_state() as an efficiency
        # optimization; is safe so long as we don't modify the wrapped state.
        state._wrapped_state,
        # pylint: enable=protected-access
    )
    # Censoring other players' hole cards
    per_player_action_view: list[str] = []
    for action in hand_history.actions:
      # If not dealing a hole card no need to censor.
      if not action.startswith("d dh p"):
        per_player_action_view.append(action)
        continue
      assert action.startswith("d dh p")

      # Dealing with hole cards, but still current player, so no need to censor.
      if action.startswith(f"d dh p{player + 1}"):
        per_player_action_view.append(action)
        continue
      assert not action.startswith(f"d dh p{player + 1}")

      # else: dealing hole cards for another player. This may (and usually will)
      # contain private hole cards, so censor it.
      # (We _could_ try to censor just *private* hole cards, i.e. 'down' cards
      # but not 'up' cards in games with them, but out of an abundance of
      # caution we instead just censor all of them to be sure.)

      action_components = action.split(" ")
      # for example: d, dh, p1 => d, dh, p1
      public_info = action_components[0:3]
      # for example:
      #  9hTs => ????
      #  9hTs AhTd => ???? ????
      #  7h6c4c3d2c => ??????????
      censored = ["?" * len(s) for s in action_components[3:]]
      per_player_action_view.append(" ".join(public_info + censored))
    return per_player_action_view

  def _observation_string(self, state, player):
    """Returns a string representation of the observation for a given player."""
    # pylint: disable=protected-access
    # Directly using for readonly access as an efficiency optimization. Could
    # also use deepcopy_wrapped_state() to ensure we don't accidentally mutate
    # the underlying state here, at the cost of performing the deepcopy every
    # time this method is called.
    wrapped_state_do_not_mutate: pokerkit.State = state._wrapped_state
    # pylint: enable=protected-access

    starting_stacks: list[int] = wrapped_state_do_not_mutate.starting_stacks
    stacks: list[int] = wrapped_state_do_not_mutate.stacks

    current_street_index: int | None = wrapped_state_do_not_mutate.street_index
    # Generator returning True IFF discard/draw has been performed for a given
    # street. For more details see:
    # https://pokerkit.readthedocs.io/en/stable/reference.html#pokerkit.state.State.draw_statuses
    draw_statuses: list[bool] = [
        s for s in wrapped_state_do_not_mutate.draw_statuses
    ]
    draw_status: bool | None = (
        None
        if current_street_index is None
        else draw_statuses[current_street_index]
    )
    pots: pokerkit.Pot = [p for p in wrapped_state_do_not_mutate.pots]
    player_next_to_act: int = wrapped_state_do_not_mutate.actor_index

    # --- NOTE: the following '_cards' on the state here are all generators, so
    # we need to wrap them all in list() to actually grab any values. ---

    # Unlike below, board cards are always public to everyone, so this is
    # striaghtforward.
    public_board_cards: list[pokerkit.Card] = list(
        wrapped_state_do_not_mutate.board_cards
    )

    # Hole cards for this current player that only this current player can see.
    private_down_cards: list[pokerkit.Card] = list(
        wrapped_state_do_not_mutate.get_down_cards(player)
    )

    # Creates a list containing 1. any Public (face-up) hole cards for player p,
    # plus '??'s for any of their facedown cards (**even if player p is the
    # current player or the 'observing' player**).
    # As such, this will look identical in all players' observations.
    per_player_censored_cards: list[pokerkit.Card] = [
        list(wrapped_state_do_not_mutate.get_censored_hole_cards(p))
        for p in range(wrapped_state_do_not_mutate.player_count)
    ]
    return (
        f"Player: {player}\n"
        f"||Current Street: {current_street_index}\n"
        f"||Current Street discard/draw performed: {draw_status}\n"
        f"||Next Player to act: {player_next_to_act}\n"
        f"||Pot(s): {pots} \n"
        f"||Board Cards: {public_board_cards}\n"
        f"||Player's Private Hole Cards: {private_down_cards}\n"
        # NOTE: This is intentionally the same for all observing players. I.e.
        # even for player {player}, the private hole cards will be ??-censored.
        f"||Per-player Hole Cards (public view): {per_player_censored_cards}\n"
        f"||Per-player Starting Stacks: {starting_stacks}\n"
        f"||Per-player Current Stacks: {stacks}"
    )

  def set_from(self, state, player) -> None:
    """No-op; see `provides_observation_tensor=False,` above.

    Defined only because some tests (incorrectly?) assume this method will
    always be defined.

    Args:
      state: The state to extract the observation from.
      player: The player to extract the observation for.
    """
    state.get_game().raise_error_if_player_out_of_range(player)
    pass

  def string_from(self, state, player) -> str:
    """An observation of `state` from the PoV of `player`, as a string.

    If the Observer's iig_obs_types `perfect_recall` is True, then this is
    equivalent to returning an `information_state_string`. If `perfect_recall`
    is False, then this is equivalent to returning an `observation_string`.

    Args:
      state: The state to extract the observation from.
      player: The player to extract the observation for.

    Returns:
      A string representation of the observation.
    """
    observation_string = self._observation_string(state, player)
    if self.iig_obs_types.perfect_recall:
      # I.e. construct the 'information state string' (by simply merging the
      # observation string with the PHH actions).
      phh_action_string = ",".join(
          self._poker_hand_history_actions(state, player)
      )
      return f"{observation_string}\n||PHH Actions: {phh_action_string}"
    else:
      return self._observation_string(state, player)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# TODO: b/434776281 - extract out PokerkitWrapperAcpcStyle into a separate file.


# --- "ACPC-style" PokerkitWrapper subclass to mimick UniversalPoker ---
class PokerkitWrapperAcpcStyle(PokerkitWrapper):
  """Subclass that mimics ACPC-wrapping UniversalPoker action handling.

  NOTE: To compare with UniversalPoker, please use the PHH hand history and/or
  'returns' array as your source of truth.

  - only action handling (applying actions, listing legal actions, converting
    actions to strings) is modified. The info/observation tensors and strings
    will exactly match that of an un-subclassed PokerkitWrapper - and therefore
    will NOT match UniversalPoker.

  - action_to_string's only change is to interpet the provided
    action as an ACPC-style action; it will simply map the action back to a
    pokerkit style action before returning the string as per usual (i.e. it will
    not make any other changes to the output string, e.g. to match
    universal_poker).
  """

  def __init__(self, params=None):
    if (
        params
        and "variant" in params
        and params["variant"] not in _VARIANTS_SUPPORTING_ACPC_STYLE
    ):
      raise ValueError(
          "PokerkitWrapperAcpcStyle only supports 'limit_holdem' and "
          "'nolimit_holdem' variants."
      )
    # Note: on _game_type() to ensure that we use the correct game type in the
    # base class's init().
    super().__init__(params)

  def _game_type(self):
    """Returns the game type of the underlying game.

    Needed to ensure that inside the super().__init()__ we use the correct
    one (not the base class's).
    """
    return _GAME_TYPE_ACPC_STYLE

  def _calculate_max_game_length(self):
    """Estimate maximum game like UniversalPokerGame::MaxGameLength()."""
    if self._max_game_length_estimate is not None:
      return self._max_game_length_estimate

    dummy_wrapped_state = self.wrapped_state_factory()
    num_players = dummy_wrapped_state.player_count
    length = 1
    streets = dummy_wrapped_state.streets
    for street in streets:
      length += street.board_dealing_count
      length += len(street.hole_dealing_statuses) * num_players
    length += len(streets) * (num_players)
    max_stack = max(self.stack_sizes)
    max_blind = max(self.blinds)
    max_num_raises = math.ceil((max_stack + max_blind - 1) / max_blind)
    length += max_num_raises * (num_players - 1)
    return length

  def game_type(self):
    return _GAME_TYPE_ACPC_STYLE

  def new_initial_state(self):
    return PokerkitWrapperAcpcStyleState(self)


class PokerkitWrapperAcpcStyleState(PokerkitWrapperState):
  """Represents an _OpenSpiel_ 'state' for the PokerkitWrapper game.

  As described by the name, this class indeed wraps a `pokerkit.State` object
  and provides the necessary interface for OpenSpiel's `pyspiel.State`.
  """

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    current_payoffs = self._wrapped_state.payoffs
    self._action_converter = ToAcpcActionConverter(
        num_players=game.num_players(),
        initial_contributions={
            # Payoffs should contain the negative initial contributions from
            # blinds and/or straddles, antes, and/or bring-ins. Hence it's
            # easier to just use it rather than checking the relevant values for
            # the given game variant.
            p: -1 * current_payoffs[p]
            for p in range(game.num_players())
        },
    )
    self._pokerkit_action_to_acpc_style_mapping = {}
    self._acpc_action_to_pokerkit_action_mapping = {}
    self._refresh_action_mappings_if_player_node()

  def _refresh_action_mappings_if_player_node(self):
    """Refreshes the action mappings if we are at a player node.

    NOTE: should be called upon initializaiton + immediately *after* applying
    any action - not at the start of e.g. _apply_action. This is because we use
    these maps in multiple mostly-pure functions such as action_to_string() and
    _legal_actions(). Otherwise we would have to call this function at the
    start of each of those functions, i.e. regenerating the dict every time they
    are called (which would be inefficient and much less elegant).
    """
    if self.is_chance_node() or self.is_terminal():
      return

    assert not self.is_chance_node() and not self.is_terminal()
    current_player = self.current_player()
    # NOTE: not calling self._legal_actions() or self._action_to_string(). Since
    # those depends on these maps *having already been updated* + are in the
    # world of ACPC-style actions we're trying to support, instead of the
    # pokerkit-style actions we have at this point.
    pokerkit_style_actions = super()._legal_actions(current_player)
    pokerkit_style_action_strings = [
        super()._action_to_string(current_player, action)
        for action in pokerkit_style_actions
    ]
    self._pokerkit_action_to_acpc_style_mapping = (
        self._action_converter.create_action_map(
            current_player,
            pokerkit_style_actions,
            pokerkit_style_action_strings,
        )
    )
    self._acpc_action_to_pokerkit_action_mapping = {
        v: k for k, v in self._pokerkit_action_to_acpc_style_mapping.items()
    }

  def to_pokerkit_action(self, acpc_style_action):
    return self._acpc_action_to_pokerkit_action_mapping[acpc_style_action]

  def to_acpc_action(self, pokerkit_style_action):
    return self._pokerkit_action_to_acpc_style_mapping[pokerkit_style_action]

  def _action_to_string(self, player, action):
    """Returns the string representation of an ACPC style action.

    NOTE: Maps ACPC Style actions to a Pokerkit Style actions, then returns the
    base class's _action_to_string(). Does NOT attempt to make the output string
    exactly match that of UniversalPoker.

    Args:
      player: The player for whom the action string is generated.
      action: The ACPC-style action.
    """
    if self.is_chance_node():
      return super()._action_to_string(player, action)
    if self.is_terminal():
      raise ValueError("Cannot convert action to string for a terminal state.")

    if action not in self._acpc_action_to_pokerkit_action_mapping:
      assert action in self.legal_actions(player)
      raise ValueError(
          f"Action {action} is not a valid ACPC style action. Expected legal"
          f" actions are: {self.legal_actions(player)}"
      )
    assert not (self.is_chance_node() or self.is_terminal())
    assert action in self._acpc_action_to_pokerkit_action_mapping
    return super()._action_to_string(
        player, self._acpc_action_to_pokerkit_action_mapping[action]
    )

  def _legal_actions(self, player: int) -> list[int]:
    """Returns the legal actions in ACPC style.

    Obtains the list of pokerkit-style actions from the base class, then
    converts them all to ACPC style.

    Args:
      player: The player for whom to get the legal actions.

    Returns:
      A list of legal actions in ACPC style.
    """
    pokerkit_style_legal_actions = super()._legal_actions(player)
    if self.is_chance_node() or self.is_terminal():
      return pokerkit_style_legal_actions
    else:
      return [self.to_acpc_action(a) for a in pokerkit_style_legal_actions]

  def _apply_action(self, action):
    """Apply ACPC-style actions.

    Aplies ACPC actions by converting them to Pokerkit style actions and then
    passing them to the base class's _apply_action(). Also perorms various
    validation checks and bookkeeping.

    Args:
      action: The ACPC-style action to apply.
    """
    if self.is_chance_node():
      self._action_converter.track_chance_node()
      super()._apply_action(action)

      # Refresh so that the *next* action will be handled correctly.
      self._refresh_action_mappings_if_player_node()
      return

    if self.is_terminal():
      raise ValueError("Cannot apply action to a terminal state.")

    assert not (self.is_chance_node() or self.is_terminal())
    pokerkit_style_action = self._acpc_action_to_pokerkit_action_mapping[action]

    # Avoid bugs due to accidentally using the unconverted action (which is
    # ACPC-style still at this point)
    acpc_style_action = action
    del action
    if pokerkit_style_action == ACTION_FOLD:
      super()._apply_action(pokerkit_style_action)
      self._refresh_action_mappings_if_player_node()
      return

    if pokerkit_style_action == ACTION_CHECK_OR_CALL:
      size = self._wrapped_state.checking_or_calling_amount
    elif pokerkit_style_action > ACTION_CHECK_OR_CALL:
      # Handling completion bet or raise-to. *Usually* the action is the size,
      # but there are rare exceptions.
      size = pokerkit_style_action
      # NOTE: the first call is using super() meaning it will interpret the
      # action as a *pokerkit style* action! Not an ACPC style action, like we
      # would get it using self.action_to_string().
      if (
          not super()
          ._action_to_string(self.current_player(), pokerkit_style_action)
          .endswith(f"Bet/Raise to {size}")
      ):
        # Currently there is only one very specific edge case where this is
        # allowed: betting exactly one chip.
        assert (
            super()
            ._action_to_string(self.current_player(), pokerkit_style_action)
            .endswith("Bet/Raise to 1 [ALL-IN EDGECASE]")
        )
        size = 1
      # NOTE: now double-checking using the *subclassed* action_to_string()
      # method, hence passing in the acpc_style_action.
      assert self._action_to_string(
          self.current_player(), acpc_style_action
      ).endswith(f"Bet/Raise to {size}")
    else:
      raise ValueError(
          "Action must be a check, call, completion bet, or raise-to."
      )
    assert size is not None

    self._action_converter.track_pokerkit_style_player_action(
        self.current_player(), pokerkit_style_action, size
    )
    super()._apply_action(pokerkit_style_action)
    self._refresh_action_mappings_if_player_node()
    return


class ToAcpcActionConverter:
  """Helper class to convert PokerKit-style actions to ACPC-style actions.

  Specifically: PokerKit's betting actions generally mean "raise this player's
  contribution for *this* street to X". This differs from ACPC's betting actions
  which generally represent "raise this player's toatal contribution across all
  streets to X". This class helps track contributions to allow us to convert
  between these two styles as needed in PokerKitWrapperACPCStyle.
  """

  def __init__(
      self,
      num_players: int,
      initial_contributions: dict[int, float],
  ):
    if num_players is None:
      raise ValueError("num_players must be specified.")
    if num_players <= 1:
      raise ValueError("num_players must be at least 2.")
    if initial_contributions is None:
      raise ValueError("initial_contributions must be specified.")

    self._num_players = num_players
    self._per_street_contributions: list[dict[int, float]] = [
        {p: 0.0 for p in range(self._num_players)} | initial_contributions
    ]

    for p in range(num_players):
      assert p in self._per_street_contributions[0]
    if len(self._per_street_contributions[0]) != num_players:
      raise ValueError(
          "After merging in initial_contributions, expected"
          f" {num_players} players. Instead got"
          f" {len(initial_contributions)} players."
      )

    # NOTE: Defaults to False since we also default self.current_street to 0 and
    # don't want track_chance_node() to increment the street
    # counter if people call it prior to any player actions, e.g. dealing cards
    # in Hold'em preflop chance nodes.
    self._ready_to_increment_street = False

  def track_chance_node(self):
    """Handles bookkeeping related to chance nodes potentially changing street.

    Specifically: increments the street IFF we have not tracked any additional
    contributions since the last time we incremented the street.

    In practice, users can call this function every time they are processing
    a chance node (and trust it will increment only once to the next street, on
    the first such call).

    Raises:
      RuntimeError: If the street is incremented past the maximum number of
        streets.
    """
    if not self._ready_to_increment_street:
      # E.g. called a second or third time when dealing the 3 flop cards.
      return
    # else:
    self._ready_to_increment_street = False
    self._per_street_contributions.append(
        {p: 0.0 for p in range(self._num_players)}
    )

  def track_pokerkit_style_player_action(
      self, player, pokerkit_style_action, size
  ):
    """Tracks a completion bet or raise, or check or call, for the given player.

    Stores information for the given action **assuming said action is NOT in
    ACPC stlye**, and marks that the street is ready to be incremented upon the
    next chance node.

    (If you want to track an action that is in ACPC style, please first convert
    it to the non-ACPC 'normal' PokerkitWrapper action before calling this
    method!)

    NOTE: Size should not necessarily equal the pokerkit_style_action (even if
    it does for most bets). In particular:
    - Size should be 0 for checks
    - Size should equal the 'calling amount' for calls.
    - Size should equal the amount completion-bet or raised-to. (This is
      _usually_ equal to the action, but there are exceptions: e.g. in
      situations where betting 1 chip is allowed.)

    Args:
      player: The index of the player who performed the action.
      pokerkit_style_action: The action taken by the player, using the
        PokerkitWrapper's action encoding (e.g., ACTION_FOLD,
        ACTION_CHECK_OR_CALL, or a bet/raise amount).
      size: The numerical value associated with the action. This is 0 for folds
        and checks, the calling amount for calls, and the total amount being
        bet/raised to for completion bets or raises.
    """
    if size < 0:
      raise ValueError("Size must be non-negative.")

    # The caller may still have additional contributions on this street, but we
    # can at least guarantee that on the *next* chance node it will be a
    # different street.
    self._ready_to_increment_street = True
    if pokerkit_style_action in FOLD_AND_CHECK_OR_CALL_ACTIONS and size == 0:
      pass  # Fold and check contribute no additional chips from their stack.
      return
    elif pokerkit_style_action == ACTION_FOLD:  # and size != 0
      raise ValueError("FOLD actions should be passed as 0 size.")
    elif pokerkit_style_action == ACTION_CHECK_OR_CALL:  # and size != 0
      # This MUST be a call if we've reacehd this point, since checks
      # (i.e. size == 0) should have been handled above already above.
      assert(size > 0)
      # NOTE: Adding, not replacing, since call sizes in pokerkit are defined
      # as the *additional* contribution on the current street (unlike
      # completion bets or raises).
      self._per_street_contributions[-1][player] += size
    elif pokerkit_style_action > max(FOLD_AND_CHECK_OR_CALL_ACTIONS):
      if pokerkit_style_action != size and not (
          pokerkit_style_action == 2 and size == 1
      ):
        raise ValueError(
            "pokerkit_style_action must equal size for completion bets or"
            " raise-tos (except very special cases whereaction 2 can mean 'bet"
            f" 1 chip'). Got action {pokerkit_style_action} and size"
            f" {size} instead."
        )
      # Completion bet or raise-to. (Note how unlike calls, this is effectively
      # "replacing" the contribution for the current street, not just adding to
      # it.)
      self._per_street_contributions[-1][player] = size
    else:
      raise ValueError(
          "pokerkit_style_action must be a fold, check, or call. Got action"
          f" {pokerkit_style_action} instead."
      )

  def create_action_map(
      self,
      player: int,
      pokerkit_style_actions: list[int],
      pokerkit_style_action_strings: list[str],
  ) -> dict[int, int]:
    """Creates an dict mapping Pokerkit-style actions to Acpc-style actions."""
    if len(pokerkit_style_actions) != len(pokerkit_style_action_strings):
      raise ValueError(
          "pokerkit_style_actions and action_strings must have the same length."
      )
    pokerkit_action_and_string_pairs = list(
        zip(pokerkit_style_actions, pokerkit_style_action_strings)
    )

    for (
        pokerkit_style_action,
        pokerkit_style_action_string,
    ) in pokerkit_action_and_string_pairs:
      if pokerkit_style_action not in FOLD_AND_CHECK_OR_CALL_ACTIONS:
        expected_bet_raise_string = "Bet/Raise to"
        if expected_bet_raise_string not in pokerkit_style_action_string:
          raise ValueError(
              f"Action string does not contain '{expected_bet_raise_string}'."
              f" Got action string: {pokerkit_style_action_string}"
          )

    # This is the main purpose for having this separate function: it allows us
    # to calculate the total prior street contribution once, rather than having
    # to do so for every single action.
    current_street = len(self._per_street_contributions) - 1
    # NOTE: exclusive slice to avoid including the current street's
    # contributions in the sum here.
    prior_street_contributions = self._per_street_contributions[:current_street]
    total_prior_street_contribution = int(sum(
        contribution[player] for contribution in prior_street_contributions
    ))
    pokerkit_action_to_acpc_action_mapping = {
        pokerkit_style_action: self._convert_action_to_acpc_style(
            pokerkit_style_action,
            pokerkit_style_action_string,
            total_prior_street_contribution,
        )
        for pokerkit_style_action, pokerkit_style_action_string in pokerkit_action_and_string_pairs
    }
    # All our actions should be unique by definition. If not, there's likely a
    # but in our logic.
    assert len(pokerkit_action_to_acpc_action_mapping.keys()) == len(
        set(pokerkit_action_to_acpc_action_mapping.keys())
    )
    assert len(pokerkit_action_to_acpc_action_mapping.values()) == len(
        set(pokerkit_action_to_acpc_action_mapping.values())
    )
    return pokerkit_action_to_acpc_action_mapping

  def _convert_action_to_acpc_style(
      self,
      pokerkit_style_action: int,
      pokerkit_style_action_string: str,
      total_prior_street_contribution: float,
  ) -> int:
    """Helper convert indvidiual actions for convert_pokerkit_style_actions."""
    if pokerkit_style_action in FOLD_AND_CHECK_OR_CALL_ACTIONS:
      # Fold and check_or_call actions are always hardcoded to the same values
      # in ACPC style and pokerkit style, so no need to convert the action in
      # any way.
      return pokerkit_style_action

    # Completion Bet or Raise-to
    #
    # NOTE: This is the one place where we usually have to convert actions!
    # I.e. ACPC-style and pokerkit-style actions primarily differ only in this
    # specific point.
    pokerkit_action_size = pokerkit_style_action

    # Handle a very specific edge-case where the player is allowed to bet
    # exactly one chip via a special '2' action (to avoid conflict with
    # the ACTION_CHECK_OR_CALL value of 1).
    if pokerkit_style_action == 2 and pokerkit_style_action_string.endswith(
        "Bet/Raise to 1 [ALL-IN EDGECASE]"
    ):
      pokerkit_action_size = 1
    assert (
        f"Bet/Raise to {pokerkit_action_size}" in pokerkit_style_action_string
    )
    return pokerkit_action_size + total_prior_street_contribution


# ------------------------------------------------------------------------------

pyspiel.register_game(_GAME_TYPE, PokerkitWrapper)
pyspiel.register_game(_GAME_TYPE_ACPC_STYLE, PokerkitWrapperAcpcStyle)
